import pytest

from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
)


@pytest.fixture(scope="function")
def one_owner_and_two_developers(gitlab, group, users):

    gitlab.add_member_to_group(group, users[0], AccessLevel.OWNER.value)
    gitlab.add_member_to_group(group, users[1], AccessLevel.DEVELOPER.value)
    gitlab.add_member_to_group(group, users[2], AccessLevel.DEVELOPER.value)
    gitlab.remove_member_from_group(group, "root")

    yield group

    # we are running tests with root's token, so every group is created
    # with a single user - root as owner. we restore the group to
    # this state here.

    gitlab.add_member_to_group(group, "root", AccessLevel.OWNER.value)

    # we try to remove all users, not just those added above,
    # on purpose, as more may have been added in the tests
    for user in users:
        gitlab.remove_member_from_group(group, user)


class TestGroupMembersUsers:
    def test__add_user(self, gitlab, group, users, one_owner_and_two_developers):

        no_of_members_before = len(gitlab.get_group_members(group))
        user_to_add = f"{users[3]}"

        add_users = f"""
            projects_and_groups:
              {group}/*:
                group_members:
                  users:
                    {users[0]}:
                      access_level: {AccessLevel.OWNER.value}
                    {users[1]}:
                      access_level: {AccessLevel.DEVELOPER.value}
                    {users[2]}:
                      access_level: {AccessLevel.DEVELOPER.value}
                    {user_to_add}: # new user 1
                      access_level: {AccessLevel.DEVELOPER.value}
            """

        run_gitlabform(add_users, group)

        members = gitlab.get_group_members(group)
        assert len(members) == no_of_members_before + 1

        members_usernames = [member["username"] for member in members]
        assert members_usernames.count(user_to_add) == 1

    def test__remove_user(self, gitlab, group, users, one_owner_and_two_developers):

        no_of_members_before = len(gitlab.get_group_members(group))
        user_to_remove = f"{users[2]}"

        remove_users = f"""
            projects_and_groups:
              {group}/*:
                group_members:
                  users:
                    {users[0]}:
                      access_level: {AccessLevel.OWNER.value}
                    {users[1]}:
                      access_level: {AccessLevel.DEVELOPER.value}
                  enforce: tru
            """

        run_gitlabform(remove_users, group)

        members = gitlab.get_group_members(group)
        assert len(members) == no_of_members_before - 1

        members_usernames = [member["username"] for member in members]
        assert user_to_remove not in members_usernames

    def test__not_remove_users_with_enforce_false(
        self, gitlab, group, users, one_owner_and_two_developers
    ):

        no_of_members_before = len(gitlab.get_group_members(group))

        setups = [
            # flag explicitly set to false
            f"""
                projects_and_groups:
                  {group}/*:
                    group_members:
                      users:
                        {users[0]}:
                          access_level: {AccessLevel.OWNER.value}
                        {users[1]}:
                          access_level: {AccessLevel.DEVELOPER.value}
                      enforce: false
                """,
            # flag not set at all (but the default is false)
            f"""
                projects_and_groups:
                  {group}/*:
                    group_members:
                      users:
                        {users[0]}:
                          access_level: {AccessLevel.OWNER.value}
                        {users[1]}:
                          access_level: {AccessLevel.DEVELOPER.value}
                """,
        ]
        for setup in setups:
            run_gitlabform(setup, group)

            members = gitlab.get_group_members(group)
            assert len(members) == no_of_members_before

            members_usernames = {member["username"] for member in members}
            assert members_usernames == {
                f"{users[0]}",
                f"{users[1]}",
                f"{users[2]}",
            }

    def test__change_some_users_access(
        self, gitlab, group, users, one_owner_and_two_developers
    ):

        new_access_level = AccessLevel.MAINTAINER.value

        change_some_users_access = f"""
            projects_and_groups:
              {group}/*:
                group_members:
                  users:
                    {users[0]}:
                      access_level: {AccessLevel.OWNER.value}
                    {users[1]}:
                      access_level: {new_access_level} # changed level
                    {users[2]}:
                      access_level: {new_access_level} # changed level
            """

        run_gitlabform(change_some_users_access, group)

        members = gitlab.get_group_members(group)
        for member in members:
            if member["username"] == f"{users[0]}":
                assert member["access_level"] == AccessLevel.OWNER.value
            if member["username"] == f"{users[1]}":
                assert member["access_level"] == new_access_level
            if member["username"] == f"{users[2]}":
                assert member["access_level"] == new_access_level

    def test__change_owner(self, gitlab, group, users, one_owner_and_two_developers):

        change_owner = f"""
            projects_and_groups:
              {group}/*:
                group_members:
                  users:
                    {users[0]}:
                      access_level: {AccessLevel.DEVELOPER.value} # only developer now
                    {users[1]}:
                      access_level: {AccessLevel.OWNER.value} # new owner
                    {users[2]}:
                      access_level: {AccessLevel.DEVELOPER.value}
            """

        run_gitlabform(change_owner, group)

        members = gitlab.get_group_members(group)
        assert len(members) == 3

        for member in members:
            if member["username"] == f"{users[0]}":
                assert (
                    member["access_level"] == AccessLevel.DEVELOPER.value
                )  # only developer now
            if member["username"] == f"{users[1]}":
                assert member["access_level"] == AccessLevel.OWNER.value  # new owner
            if member["username"] == f"{users[2]}":
                assert member["access_level"] == AccessLevel.DEVELOPER.value

    def test__add_user_with_access_level_name(
        self, gitlab, group, users, one_owner_and_two_developers
    ):

        no_of_members_before = len(gitlab.get_group_members(group))
        user_to_add = f"{users[3]}"

        add_users = f"""
        projects_and_groups:
          {group}/*:
            group_members:
              {users[0]}:
                access_level: owner
              {users[1]}:
                access_level: developer
              {users[2]}:
                access_level: developer
              {user_to_add}: # new user 1
                access_level: developer
        """

        run_gitlabform(add_users, group)

        members = gitlab.get_group_members(group)
        assert len(members) == no_of_members_before + 1

        members_usernames = [member["username"] for member in members]
        assert members_usernames.count(user_to_add) == 1



import pytest

from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
)


@pytest.fixture(scope="function")
def tags(request, gitlab, group_and_project):
    tags = [
        "tag1",
        "tag2",
        "tag3",
    ]
    for tag in tags:
        gitlab.create_tag(group_and_project, tag, "main")

    def fin():
        protected_tags = gitlab.get_protected_tags(group_and_project)
        for protected_tag in protected_tags:
            gitlab.unprotect_tag(group_and_project, protected_tag["name"])
        for tag in tags:
            gitlab.delete_tag(group_and_project, tag)

    request.addfinalizer(fin)


class TestTags:
    def test__protect_single_tag(self, gitlab, group_and_project, tags):

        config = f"""
        projects_and_groups:
          {group_and_project}:
            tags:
              tag1:
                protected: true
                create_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config, group_and_project)

        tags = gitlab.get_tags(group_and_project)
        for tag in tags:
            if tag["name"] == "tag1":
                assert tag["protected"]
            else:
                assert not tag["protected"]

        protected_tags = gitlab.get_protected_tags(group_and_project)
        assert len(protected_tags) == 1
        assert protected_tags[0]["name"] == "tag1"
        assert (
            protected_tags[0]["create_access_levels"][0]["access_level"]
            == AccessLevel.MAINTAINER.value
        )

    def test__protect_wildcard_tag(self, gitlab, group_and_project, tags):

        config = f"""
        projects_and_groups:
          {group_and_project}:
            tags:
              "tag*":
                protected: true
                create_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config, group_and_project)

        tags = gitlab.get_tags(group_and_project)
        for tag in tags:
            assert tag["protected"]

        protected_tags = gitlab.get_protected_tags(group_and_project)
        assert len(protected_tags) == 1
        assert protected_tags[0]["name"] == "tag*"
        assert (
            protected_tags[0]["create_access_levels"][0]["access_level"]
            == AccessLevel.MAINTAINER.value
        )

    def test__unprotect_the_same_tag(self, gitlab, group_and_project, tags):

        config = f"""
        projects_and_groups:
          {group_and_project}:
            tags:
              "tag*":
                protected: true
                create_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config, group_and_project)

        tags = gitlab.get_tags(group_and_project)
        for tag in tags:
            assert tag["protected"]

        protected_tags = gitlab.get_protected_tags(group_and_project)
        assert len(protected_tags) == 1
        assert protected_tags[0]["name"] == "tag*"
        assert (
            protected_tags[0]["create_access_levels"][0]["access_level"]
            == AccessLevel.MAINTAINER.value
        )

        config = f"""
        projects_and_groups:
          {group_and_project}:
            tags:
              "tag*":
                protected: false
        """

        run_gitlabform(config, group_and_project)

        tags = gitlab.get_tags(group_and_project)
        for tag in tags:
            assert not tag["protected"]

        protected_tags = gitlab.get_protected_tags(group_and_project)
        assert len(protected_tags) == 0

    def test__protect_single_tag_no_access(self, gitlab, group_and_project, tags):

        config = f"""
            projects_and_groups:
              {group_and_project}:
                tags:
                  tag1:
                    protected: true
                    create_access_level: {AccessLevel.NO_ACCESS.value}
            """

        run_gitlabform(config, group_and_project)

        tags = gitlab.get_tags(group_and_project)
        for tag in tags:
            if tag["name"] == "tag1":
                assert tag["protected"]
            else:
                assert not tag["protected"]

        protected_tags = gitlab.get_protected_tags(group_and_project)
        assert len(protected_tags) == 1
        assert protected_tags[0]["name"] == "tag1"
        assert (
            protected_tags[0]["create_access_levels"][0]["access_level"]
            == AccessLevel.NO_ACCESS.value
        )

    def test__protect_single_tag_with_access_level_names(
        self, gitlab, group_and_project, tags
    ):
        config = f"""
            projects_and_groups:
              {group_and_project}:
                tags:
                  tag1:
                    protected: true
                    create_access_level: maintainer
            """

        run_gitlabform(config, group_and_project)

        tags = gitlab.get_tags(group_and_project)
        for tag in tags:
            if tag["name"] == "tag1":
                assert tag["protected"]
            else:
                assert not tag["protected"]

        protected_tags = gitlab.get_protected_tags(group_and_project)
        assert len(protected_tags) == 1
        assert protected_tags[0]["name"] == "tag1"
        assert (
            protected_tags[0]["create_access_levels"][0]["access_level"]
            == AccessLevel.MAINTAINER.value
        )



from tests.acceptance import run_gitlabform


class TestProjectSettings:
    def test__builds_for_private_projects(self, gitlab, group_and_project):
        settings = gitlab.get_project_settings(group_and_project)
        assert settings["visibility"] == "private"

        config_builds_for_private_projects = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              visibility: internal
        """

        run_gitlabform(config_builds_for_private_projects, group_and_project)

        settings = gitlab.get_project_settings(group_and_project)
        assert settings["visibility"] == "internal"



from gitlabform.configuration.common import ConfigurationCommon
from gitlabform.configuration.groups import ConfigurationGroups
from gitlabform.configuration.projects import ConfigurationProjects


#
# This the only external interface for operating on the configuration from a given YAML file
# (or an input string in case of tests).
#


class Configuration(ConfigurationProjects):
    pass



from gitlabform.gitlab.core import GitLabCore


class GitLabPipelineSchedules(GitLabCore):
    def get_all_pipeline_schedules(self, project_and_group_name):
        return self._make_requests_to_api(
            "projects/%s/pipeline_schedules", project_and_group_name
        )

    def get_pipeline_schedule(self, project_and_group_name, pipeline_schedule_id):
        return self._make_requests_to_api(
            "projects/%s/pipeline_schedules/%s",
            (project_and_group_name, pipeline_schedule_id),
        )

    def create_pipeline_schedule(
        self, project_and_group_name, description, ref, cron, optional_data=None
    ):
        if optional_data is None:
            optional_data = {}
        data_required = {
            "description": description,
            "ref": ref,
            "cron": cron,
        }
        data = {**optional_data, **data_required}

        return self._make_requests_to_api(
            "projects/%s/pipeline_schedules",
            project_and_group_name,
            method="POST",
            data=data,
            expected_codes=201,
        )

    def update_pipeline_schedule(
        self, project_and_group_name, pipeline_schedule_id, data
    ):
        return self._make_requests_to_api(
            "projects/%s/pipeline_schedules/%s",
            (project_and_group_name, pipeline_schedule_id),
            method="PUT",
            data=data,
            expected_codes=[200, 201],
        )

    def take_ownership(self, project_and_group_name, pipeline_schedule_id):
        self._make_requests_to_api(
            "projects/%s/pipeline_schedules/%s/take_ownership",
            (project_and_group_name, pipeline_schedule_id),
            method="POST",
            expected_codes=[
                200,
                201,
                403,  # TODO: stop accepting 403 after GitLab idempotency bug is fixed
            ],
        )

    def delete_pipeline_schedule(self, project_and_group_name, pipeline_schedule_id):
        self._make_requests_to_api(
            "projects/%s/pipeline_schedules/%s",
            (project_and_group_name, pipeline_schedule_id),
            method="DELETE",
            expected_codes=[200, 201, 204, 404],
        )

    def create_pipeline_schedule_variable(
        self,
        project_and_group_name,
        pipeline_schedule_id,
        variable_key,
        variable_value,
        optional_data=None,
    ):
        if optional_data is None:
            optional_data = {}
        data_required = {
            "key": variable_key,
            "value": variable_value,
        }
        data = {**optional_data, **data_required}

        return self._make_requests_to_api(
            "projects/%s/pipeline_schedules/%s/variables",
            (project_and_group_name, pipeline_schedule_id),
            method="POST",
            data=data,
            expected_codes=201,
        )

    def update_pipeline_schedule_variable(
        self,
        project_and_group_name,
        pipeline_schedule_id,
        variable_key,
        variable_value,
        optional_data=None,
    ):
        if optional_data is None:
            optional_data = {}
        data_required = {
            "value": variable_value,
        }
        data = {**optional_data, **data_required}

        return self._make_requests_to_api(
            "projects/%s/pipeline_schedules/%s/variables/%s",
            (project_and_group_name, pipeline_schedule_id, variable_key),
            method="PUT",
            data=data,
            expected_codes=[200, 201],
        )

    def delete_pipeline_schedule_variable(
        self, project_and_group_name, pipeline_schedule_id, variable_key
    ):
        return self._make_requests_to_api(
            "projects/%s/pipeline_schedules/%s/variables/%s",
            (project_and_group_name, pipeline_schedule_id, variable_key),
            method="DELETE",
            expected_codes=[200, 201, 202, 204, 404],
        )



from tests.acceptance import run_gitlabform


class TestGroupVariables:
    def test__single_variable(self, gitlab, group_for_function):
        config_single_variable = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123
        """

        run_gitlabform(config_single_variable, group_for_function)

        variables = gitlab.get_group_variables(group_for_function)
        assert len(variables) == 1

    def test__delete_variable(self, gitlab, group_for_function):
        config_single_variable = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123
        """

        run_gitlabform(config_single_variable, group_for_function)

        variables = gitlab.get_group_variables(group_for_function)
        assert len(variables) == 1
        assert variables[0]["value"] == "123"

        config_delete_variable = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123
                delete: true
        """

        run_gitlabform(config_delete_variable, group_for_function)

        variables = gitlab.get_group_variables(group_for_function)
        assert len(variables) == 0

    def test__reset_single_variable(self, gitlab, group_for_function):
        config_single_variable = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123
        """

        run_gitlabform(config_single_variable, group_for_function)

        variables = gitlab.get_group_variables(group_for_function)
        assert len(variables) == 1
        assert variables[0]["value"] == "123"

        config_single_variable2 = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123456
        """

        run_gitlabform(config_single_variable2, group_for_function)

        variables = gitlab.get_group_variables(group_for_function)
        assert len(variables) == 1
        assert variables[0]["value"] == "123456"

    def test__more_variables(self, gitlab, group_for_function):
        config_more_variables = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123456
              bar:
                key: BAR
                value: bleble
        """

        run_gitlabform(config_more_variables, group_for_function)

        variables = gitlab.get_group_variables(group_for_function)
        variables_keys = {variable["key"] for variable in variables}
        assert len(variables) == 2
        assert variables_keys == {"FOO", "BAR"}

    def test__masked_variables(self, gitlab, group_for_function):
        masked_variables = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                # https://docs.gitlab.com/ee/ci/variables/#masked-variable-requirements
                value: 12345678
                masked: true
        """

        run_gitlabform(masked_variables, group_for_function)

        variable = gitlab.get_group_variable_object(group_for_function, "FOO")
        assert variable["value"] == "12345678"
        assert variable["masked"]

    def test__protected_variables(self, gitlab, group_for_function):
        protected_variables = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123
                protected: true
        """

        run_gitlabform(protected_variables, group_for_function)

        variable = gitlab.get_group_variable_object(group_for_function, "FOO")
        assert variable["value"] == "123"
        assert variable["protected"]

    def test__protected_change_variables(self, gitlab, group_for_function):
        config_single_variable = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123
        """

        run_gitlabform(config_single_variable, group_for_function)

        variable = gitlab.get_group_variable_object(group_for_function, "FOO")
        assert variable["value"] == "123"
        assert variable["protected"] is False

        protected_variables = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: FOO
                value: 123
                protected: true
        """

        run_gitlabform(protected_variables, group_for_function)

        variable = gitlab.get_group_variable_object(group_for_function, "FOO")
        assert variable["value"] == "123"
        assert variable["protected"] is True

    def test__not_masked_and_not_protected_variable(self, gitlab, group_for_function):
        config_single_variable = f"""
        projects_and_groups:
          {group_for_function}/*:
            group_variables:
              foo:
                key: DOUBLE_NOT
                value: 123
                masked: false
                protected: false
        """

        run_gitlabform(config_single_variable, group_for_function)

        variable = gitlab.get_group_variable_object(group_for_function, "DOUBLE_NOT")
        assert variable["masked"] is False
        assert variable["protected"] is False



from gitlabform.gitlab.projects import GitLabProjects


class GitLabProjectBadges(GitLabProjects):
    def get_project_badges(self, project_and_group_name):
        badges = self._make_requests_to_api(
            "projects/%s/badges",
            project_and_group_name,
            expected_codes=200,
        )
        # according to the docs_new this endpoint returns also the group badges
        # but we want only the project badges here
        return [badge for badge in badges if badge["kind"] == "project"]

    def add_project_badge(
        self,
        project_and_group_name,
        badge_in_config,
    ):
        return self._make_requests_to_api(
            "projects/%s/badges",
            project_and_group_name,
            method="POST",
            data=badge_in_config,
            expected_codes=201,
        )

    def edit_project_badge(
        self,
        project_and_group_name,
        badge_in_gitlab,
        badge_in_config,
    ):
        return self._make_requests_to_api(
            "projects/%s/badges/%s",
            (project_and_group_name, badge_in_gitlab["id"]),
            method="PUT",
            data=badge_in_config,
        )

    def delete_project_badge(
        self,
        project_and_group_name,
        badge_in_gitlab,
    ):
        # 404 means it is already removed, so let's accept it for idempotency
        return self._make_requests_to_api(
            "projects/%s/badges/%s",
            (project_and_group_name, badge_in_gitlab["id"]),
            method="DELETE",
            expected_codes=[200, 204, 404],
        )



import functools

from gitlabform.gitlab.core import GitLabCore, NotFoundException


class GitLabGroups(GitLabCore):
    @functools.lru_cache()
    def get_group_id_case_insensitive(self, some_string):
        # Cache the mapping from some_string -> id, as that won't change during our run.
        return self.get_group_case_insensitive(some_string)["id"]

    def get_group_case_insensitive(self, some_string):

        # maybe "foo/bar" is some group's path

        try:
            # try with exact case
            return self.get_group(some_string)
        except NotFoundException:

            # try case insensitive
            groups = self._make_requests_to_api(
                "groups?search=%s",
                some_string.lower(),
                method="GET",
            )

            for group in groups:
                if group["full_path"].lower() == some_string.lower():
                    return group
            raise NotFoundException(
                f"Group/subgroup with path '{some_string}' not found."
            )

    def create_group(self, name, path, parent_id=None, visibility="private"):
        data = {
            "name": name,
            "path": path,
            "visibility": visibility,
        }
        if parent_id is not None:
            data["parent_id"] = parent_id
        return self._make_requests_to_api(
            "groups", data=data, method="POST", expected_codes=201
        )

    def delete_group(self, group_name):
        # 404 means that the group does not exist anymore, so let's accept it for idempotency
        return self._make_requests_to_api(
            "groups/%s",
            group_name,
            method="DELETE",
            expected_codes=[200, 202, 204, 404],
        )

    def get_group(self, name):
        return self._make_requests_to_api("groups/%s", name)

    def get_groups(self):
        """
        :return: sorted list of groups
        """
        result = self._make_requests_to_api("groups?all_available=true")
        return sorted(map(lambda x: x["full_path"], result))

    def get_projects(self, group, include_archived=False, only_names=True):
        """
        :param group: group name
        :param include_archived: set to True if archived projects should also be returned
        :param only_names: set to False to get the whole project objects
        :return: sorted list of strings "group/project_name". Note that only projects from "group" namespace are
                 returned, so if "group" (= members of this group) is also a member of some projects, they won't be
                 returned here.
        """
        try:
            # there are 3 states of the "archived" flag: true, false, undefined
            # we use the last 2
            if include_archived:
                query_string = "include_subgroups=true"
            else:
                query_string = "include_subgroups=true&archived=false"

            projects = self._make_requests_to_api(
                f"groups/%s/projects?{query_string}", group
            )
        except NotFoundException:
            projects = []

        project_and_groups_in_group_namespace = [
            project
            for project in projects
            if project["path_with_namespace"].startswith(group + "/")
        ]

        if only_names:
            return sorted(
                map(
                    lambda x: x["path_with_namespace"],
                    project_and_groups_in_group_namespace,
                )
            )
        else:
            return sorted(
                project_and_groups_in_group_namespace,
                key=lambda x: x["path_with_namespace"],
            )

    def get_group_settings(self, project_and_group_name):
        try:
            return self._make_requests_to_api("groups/%s", project_and_group_name)
        except NotFoundException:
            return dict()

    def put_group_settings(self, project_and_group_name, group_settings):
        # group_settings has to be like this:
        # {
        #     'setting1': value1,
        #     'setting2': value2,
        # }
        # ..as documented at: https://docs.gitlab.com/ee/api/groups.html#update-group
        self._make_requests_to_api(
            "groups/%s", project_and_group_name, "PUT", group_settings
        )

    def get_group_shared_with(self, group):
        group = self.get_group_case_insensitive(group)

        return group["shared_with_groups"]

    def add_share_to_group(
        self, group, share_with_group_name, group_access, expires_at=None
    ):
        share_with_group_id = self.get_group_id_case_insensitive(share_with_group_name)
        data = {"group_id": share_with_group_id, "expires_at": expires_at}
        if group_access is not None:
            data["group_access"] = group_access

        return self._make_requests_to_api(
            "groups/%s/share",
            group,
            method="POST",
            data=data,
            expected_codes=[200, 201],
        )

    def remove_share_from_group(self, group, share_with_group_name):
        share_with_group_id = self.get_group_id_case_insensitive(share_with_group_name)

        # 404 means that the user is already removed, so let's accept it for idempotency
        return self._make_requests_to_api(
            "groups/%s/share/%s",
            (group, share_with_group_id),
            method="DELETE",
            expected_codes=[204, 404],
        )



from gitlabform.gitlab.core import GitLabCore


class GitLabIntegrations(GitLabCore):
    def get_integration(self, project_and_group_name, integration):
        return self._make_requests_to_api(
            "projects/%s/integrations/%s", (project_and_group_name, integration)
        )

    def set_integration(self, project_and_group_name, integration, data):
        # DO NOT CHANGE BELOW json=data, it is necessary to pass data as JSON
        # for the Integrations API to work FULLY properly!
        # see https://gitlab.com/gitlab-org/gitlab/-/issues/202216 for more info.
        self._make_requests_to_api(
            "projects/%s/integrations/%s",
            (project_and_group_name, integration),
            "PUT",
            data=None,
            expected_codes=[200, 201],
            json=data,
        )

    def delete_integration(self, project_and_group_name, integration):
        self._make_requests_to_api(
            "projects/%s/integrations/%s",
            (project_and_group_name, integration),
            "DELETE",
            expected_codes=[200, 204],
        )



import logging

import pytest

from gitlabform.configuration import Configuration

logger = logging.getLogger(__name__)


@pytest.fixture
def configuration_with_yes():
    config_yaml = """
    ---
    projects_and_groups:
      some_group/*:
        project_settings:
          foo: yes          # in YAML 1.1 this should be interpreted as boolean true
    """
    return Configuration(config_string=config_yaml)


def test__default_yaml_11(configuration_with_yes):
    assert configuration_with_yes.get(
        "projects_and_groups|some_group/*|project_settings"
    ) == {"foo": True}



from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
    get_gitlab,
)


gl = get_gitlab()


class TestBranches:
    def test__protect_and_unprotect(self, gitlab, group_and_project, branch):

        config_protect_branch = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_protect_branch, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)
        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        config_unprotect_branch = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: false
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_unprotect_branch, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)
        assert push_access_levels is None
        assert merge_access_levels is None
        assert push_access_user_ids is None
        assert merge_access_user_ids is None
        assert unprotect_access_level is None

    def test__config_with_access_level_names(self, gitlab, group_and_project, branch):

        config_with_access_levels_names = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                push_access_level: no_access        # note "_" or " " and the various
                merge_access_level: Developer       # case in each line. it should not
                unprotect_access_level: MAINTAINER  # matter as we allow any case.
        """

        run_gitlabform(config_with_access_levels_names, group_and_project)

        (
            push_access_level,
            merge_access_level,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)
        assert push_access_level == [AccessLevel.NO_ACCESS.value]
        assert merge_access_level == [AccessLevel.DEVELOPER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value



from time import sleep

from gitlabform.gitlab.core import (
    GitLabCore,
    NotFoundException,
    TimeoutWaitingForDeletion,
)


class GitLabProjects(GitLabCore):
    def get_project_case_insensitive(self, some_string):

        # maybe "foo/bar" is some project's path

        try:
            # try with exact case
            return self.get_project(some_string)
        except NotFoundException:

            # try case insensitive
            projects = self._make_requests_to_api(
                f"projects?search=%s&simple=true",
                some_string.lower(),
                method="GET",
            )
            for project in projects:
                if project["path_with_namespace"].lower() == some_string.lower():
                    return project
            raise NotFoundException(f"Project with path '{some_string}' not found.")

    def create_project(
        self,
        name,
        path,
        namespace_id,
        default_branch=None,
        wait_if_still_being_deleted=False,
    ):
        data = {
            "name": name,
            "path": path,
            "namespace_id": namespace_id,
        }
        if default_branch:
            data["default_branch"] = default_branch

        if wait_if_still_being_deleted:

            # GitLab deletes the project asynchronously, it may take a few seconds.
            # So if you are creating new project with the same name as the one
            # that is still being deleted, GitLab returns code 400
            # and "The project is still being deleted.". Let's retry a few times
            # then to start creating when the deletion is done.
            # (Note: this code DOES NOT support the "Delayed Project deletion" feature
            # where the actual deletion can be postponed for days!)

            max_retries = 10
            wait_before_retry = 3
            retry = 0

            while True:
                retry += 1

                if retry > max_retries:
                    raise TimeoutWaitingForDeletion

                response = self._make_requests_to_api(
                    "projects", data=data, method="POST", expected_codes=[201, 400]
                )
                if self._is_project_still_deleted(response):
                    # wait & retry
                    sleep(wait_before_retry)
                    continue
                else:
                    return response

        else:

            return self._make_requests_to_api(
                "projects", data=data, method="POST", expected_codes=201
            )

    @staticmethod
    def _is_project_still_deleted(response):
        # check if response looks like this:
        # {'message': {'base': ['The project is still being deleted. Please try again later.'],
        # 'limit_reached': []}}
        return (
            "message" in response
            and "base" in response["message"]
            and type(response["message"]["base"]) == list
            and len(response["message"]["base"]) == 1
            and "The project is still being deleted." in response["message"]["base"][0]
        )

    def delete_project(self, project_and_group_name):

        # 404 means that the project does not exist anymore, so let's accept it for idempotency
        return self._make_requests_to_api(
            "projects/%s",
            project_and_group_name,
            method="DELETE",
            expected_codes=[202, 204, 404],
        )

        # GitLab deletes the project asynchronously, it may take a few seconds
        # BUT it doesn't return such not yet deleted project in GET calls, so
        # there is no point in checking if it actually done here. :(
        # See create_project() for the code that deals with that.

    def get_all_projects(self, include_archived=False):
        """
        :param include_archived: if the archived projects should be returned too
        :return: sorted list of ALL projects you have access to, strings like: "group/project_name"
        """
        try:
            # there are 3 states of the "archived" flag: true, false, undefined
            # we use the last 2
            if include_archived:
                query_string = "order_by=name&sort=asc"
            else:
                query_string = "order_by=name&sort=asc&archived=false"
            result = self._make_requests_to_api(f"projects?{query_string}")
            return sorted(map(lambda x: x["path_with_namespace"], result))
        except NotFoundException:
            return []

    def get_project_settings(self, project_and_group_name):
        try:
            return self._make_requests_to_api("projects/%s", project_and_group_name)
        except NotFoundException:
            return dict()

    def put_project_settings(self, project_and_group_name, project_settings):
        # project_settings has to be like this:
        # {
        #     'setting1': value1,
        #     'setting2': value2,
        # }
        # ..as documented at: https://docs.gitlab.com/ce/api/projects.html#edit-project
        return self._make_requests_to_api(
            "projects/%s",
            project_and_group_name,
            "PUT",
            data=None,
            json=project_settings,
        )

    def get_project_push_rules(self, project_and_group_name: str):
        try:
            # for this endpoint GitLab fails if project name contains ".", so lets use pid instead
            pid: str = self._get_project_id(project_and_group_name)

            return self._make_requests_to_api("projects/%s/push_rule", pid)
        except NotFoundException:
            return dict()

    def put_project_push_rules(self, project_and_group_name: str, push_rules):
        # push_rules has to be like this:
        # {
        #     'setting1': value1,
        #     'setting2': value2,
        # }
        # ..as documented at: https://docs.gitlab.com/ee/api/projects.html#edit-project-push-rule

        # for this endpoint GitLab fails if project name contains ".", so lets use pid instead
        pid: str = self._get_project_id(project_and_group_name)

        self._make_requests_to_api("projects/%s/push_rule", pid, "PUT", push_rules)

    def post_project_push_rules(self, project_and_group_name: str, push_rules):
        # push_rules has to be like this:
        # {
        #     'setting1': value1,
        #     'setting2': value2,
        # }
        # ..as documented at: https://docs.gitlab.com/ee/api/projects.html#edit-project-push-rule

        pid: str = self._get_project_id(project_and_group_name)

        self._make_requests_to_api(
            "projects/%s/push_rule", pid, "POST", push_rules, expected_codes=201
        )

    def get_hook_id(self, project_and_group_name, url):
        hooks = self._make_requests_to_api(
            "projects/%s/hooks", project_and_group_name, "GET"
        )
        for hook in hooks:
            if hook["url"] == url:
                return hook["id"]
        return None

    def delete_hook(self, project_and_group_name, hook_id):
        self._make_requests_to_api(
            "projects/%s/hooks/%s",
            (project_and_group_name, hook_id),
            "DELETE",
            expected_codes=[200, 204],
        )

    def put_hook(self, project_and_group_name, hook_id, url, data):
        data_required = {"url": url}
        data = {**data, **data_required}
        self._make_requests_to_api(
            "projects/%s/hooks/%s", (project_and_group_name, hook_id), "PUT", data
        )

    def post_hook(self, project_and_group_name, url, data):
        data_required = {"url": url}
        data = {**data, **data_required}
        self._make_requests_to_api(
            "projects/%s/hooks",
            project_and_group_name,
            "POST",
            data,
            expected_codes=201,
        )

    def get_groups_from_project(self, project_and_group_name):
        # couldn't find an API call that was giving me directly
        # the shared groups, so I'm using directly the GET /projects/:id call
        project_info = self._make_requests_to_api("projects/%s", project_and_group_name)

        # it will return {group_name: {...api info about group_name...}, ...}
        groups = {}
        for group in project_info["shared_with_groups"]:
            groups[group["group_full_path"]] = group

        return groups

    def share_with_group(
        self, project_and_group_name, group_name, group_access, expires_at
    ):
        data = {"group_id": self._get_group_id(group_name), "expires_at": expires_at}
        if group_access is not None:
            data["group_access"] = group_access
        return self._make_requests_to_api(
            "projects/%s/share",
            project_and_group_name,
            method="POST",
            data=data,
            expected_codes=201,
        )

    def unshare_with_group(self, project_and_group_name, group_name):
        # 404 means that the group already has not access, so let's accept it for idempotency
        group_id = self._get_group_id(group_name)
        return self._make_requests_to_api(
            "projects/%s/share/%s",
            (project_and_group_name, group_id),
            method="DELETE",
            expected_codes=[204, 404],
        )

    def archive(self, project_and_group_name):
        return self._make_requests_to_api(
            "projects/%s/archive",
            project_and_group_name,
            method="POST",
            expected_codes=[200, 201],
        )

    def unarchive(self, project_and_group_name):
        return self._make_requests_to_api(
            "projects/%s/unarchive",
            project_and_group_name,
            method="POST",
            expected_codes=[200, 201],
        )



from logging import debug
from cli_ui import warning, fatal

from gitlabform.constants import EXIT_PROCESSING_ERROR, EXIT_INVALID_INPUT
from gitlabform.gitlab import GitLab
from gitlabform.gitlab.core import NotFoundException


class BranchProtector:
    new_api_keys = [
        "push_access_level",
        "merge_access_level",
        "unprotect_access_level",
    ]
    extra_param_keys = [
        "allowed_to_push",
        "allowed_to_merge",
    ]

    def __init__(self, gitlab: GitLab, strict: bool):
        self.gitlab = gitlab
        self.strict = strict

    def apply_branch_protection_configuration(
        self, project_and_group, configuration, branch
    ):
        try:
            requested_configuration = configuration["branches"][branch]

            if requested_configuration.get("protected"):
                self.protect_branch(project_and_group, configuration, branch)
            else:
                self.unprotect_branch(project_and_group, branch)

        except NotFoundException:
            message = f"Branch '{branch}' not found when trying to set it as protected/unprotected!"
            if self.strict:
                fatal(
                    message,
                    exit_code=EXIT_PROCESSING_ERROR,
                )
            else:
                warning(message)

    def protect_branch(self, project_and_group, configuration, branch):
        try:
            requested_configuration = configuration["branches"][branch]

            self.validate_branch_protection_config(
                project_and_group, requested_configuration, branch
            )

            if any(
                extra_key in requested_configuration
                for extra_key in self.extra_param_keys
            ):
                for extra_param_key in self.extra_param_keys:
                    # check if an extra_param is in config, and it contains the user parameter
                    if extra_param_key in requested_configuration and any(
                        "user" in d for d in requested_configuration[extra_param_key]
                    ):
                        for extra_config in requested_configuration[extra_param_key]:
                            # loop over the array of extra param and get the user_id related to user
                            if "user" in extra_config.keys():
                                user_id = self.gitlab._get_user_id(
                                    extra_config.pop("user")
                                )
                                extra_config["user_id"] = user_id

            if self.configuration_update_needed(
                requested_configuration, project_and_group, branch
            ):
                self.do_protect_branch(
                    requested_configuration, project_and_group, branch
                )
            else:
                debug(
                    "Skipping setting branch '%s' protection configuration because it's already as requested.",
                    branch,
                )

            if "code_owner_approval_required" in requested_configuration:

                self.set_code_owner_approval_required(
                    requested_configuration, project_and_group, branch
                )

        except NotFoundException:
            message = f"Branch '{branch}' not found when trying to set it as protected/unprotected!"
            if self.strict:
                fatal(
                    message,
                    exit_code=EXIT_PROCESSING_ERROR,
                )
            else:
                warning(message)

    def unprotect_branch(self, project_and_group, branch):
        try:
            debug("Setting branch '%s' as unprotected", branch)
            self.gitlab.unprotect_branch(project_and_group, branch)

        except NotFoundException:
            message = f"Branch '{branch}' not found when trying to set it as protected/unprotected!"
            if self.strict:
                fatal(
                    message,
                    exit_code=EXIT_PROCESSING_ERROR,
                )
            else:
                warning(message)

    def validate_branch_protection_config(
        self, project_and_group, requested_configuration, branch
    ):

        # for the new API any key needs to be defined...
        if any(
            key in requested_configuration
            for key in self.new_api_keys + self.extra_param_keys
        ):
            return
        else:
            fatal(
                f"Invalid configuration for protecting branches in project '{project_and_group}',"
                f" branch '{branch}' - missing keys. Required is any of these: "
                f"{self.new_api_keys + self.extra_param_keys}",
                exit_code=EXIT_INVALID_INPUT,
            )

    def do_protect_branch(self, requested_configuration, project_and_group, branch):
        debug("Setting branch '%s' access level", branch)

        # Protected Branches API is one of those that do not support editing entities
        # (PUT is not documented for it, at least). so you need to delete existing
        # branch protection (DELETE) and recreate it (POST) to perform an update
        # (otherwise you get HTTP 409 "Protected branch 'foo' already exists")
        self.gitlab.unprotect_branch(project_and_group, branch)

        # replace in our config our custom "user" and "group" entries with supported by
        # the Protected Branches API "user_id" and "group_id"
        for extra_param_key in self.extra_param_keys:

            for element in requested_configuration.get(extra_param_key, []):

                if "user" in element.keys():
                    user_id = self.gitlab._get_user_id(element.pop("user"))
                    element["user_id"] = user_id
                elif "group" in element.keys():
                    group_id = self.gitlab._get_group_id(element.pop("group"))
                    element["group_id"] = group_id

        protect_rules = {
            key: value
            for key, value in requested_configuration.items()
            if key != "protected"
        }

        self.gitlab.protect_branch(
            project_and_group,
            branch,
            protect_rules,
        )

    def set_code_owner_approval_required(
        self, requested_configuration, project_and_group, branch
    ):
        debug(
            "Setting branch '%s' \"code owner approval required\" option",
            branch,
        )
        self.gitlab.set_branch_code_owner_approval_required(
            project_and_group,
            branch,
            requested_configuration["code_owner_approval_required"],
        )

    def configuration_update_needed(
        self, requested_configuration, project_and_group, branch
    ):
        # get current configuration of branch access level
        current = self.get_current_branch_configuration(project_and_group, branch)

        # get requested configuration of branch access level
        requested = self.get_requested_branch_configuration(requested_configuration)

        debug("Current branch '%s' access levels: %s", branch, current)
        debug("Requested branch '%s' access levels: %s", branch, requested)
        return current != requested

    def get_current_branch_configuration(self, project_and_group_name, branch):
        # from a response for a request to Protected Branches API GET request
        # (https://docs.gitlab.com/ee/api/protected_branches.html#get-a-single-protected-branch-or-wildcard-protected-branch)
        # like:
        #
        # {
        #     "id": 1,
        #     "name": "master",
        #     "push_access_levels": [
        #         {
        #             "access_level": 40,
        #             "access_level_description": "Maintainers"
        #         }
        #     ],
        #     "merge_access_levels": [
        #         {
        #             "access_level": 40,
        #             "access_level_description": "Maintainers"
        #         }
        #     ],
        #     "allow_force_push":false,
        #     "code_owner_approval_required": false
        # }
        #
        # ...gets flat tuple of the permissions to perform each kind of 3 actions (push/merge/unprotect) and
        # the other extra settings that can be applied using Protected Branches API POST request
        #
        try:
            protected_branches_response = self.gitlab.get_branch_access_levels(
                project_and_group_name, branch
            )

            return (
                *(
                    self.get_current_permissions(protected_branches_response, "push")
                ),  # tuple of 3
                *(
                    self.get_current_permissions(protected_branches_response, "merge")
                ),  # tuple of 3
                *(
                    self.get_current_permissions(
                        protected_branches_response, "unprotect"
                    )
                ),  # tuple of 3
                protected_branches_response.get("allow_force_push"),
                # code_owner_approval_required has to use PATCH request, see set_code_owner_approval_required()
            )
        except NotFoundException:
            return tuple([None] * 10)  # = 3 * 3 + 1

    def get_current_permissions(self, protected_branches_response, action):
        # from a response for a request to Protected Branches API GET request
        # (https://docs.gitlab.com/ee/api/protected_branches.html#get-a-single-protected-branch-or-wildcard-protected-branch)
        # like:
        #
        # {
        #     "id": 1,
        #     "name": "master",
        #     "push_access_levels": [
        #         {
        #             "access_level": 40,
        #             "access_level_description": "Maintainers"
        #         }
        #     ],
        #     "merge_access_levels": [
        #         {
        #             "access_level": 40,
        #             "access_level_description": "Maintainers"
        #         }
        #     ],
        #     "allow_force_push":false,
        #     "code_owner_approval_required": false
        # }
        #
        # ...gets a tuple of sorted lists, for a given permission in ["push", "merge", "unprotect"],
        # for example for "push":
        #
        # ([40], [], [])
        #

        levels = set()
        user_ids = set()
        group_ids = set()
        # we ignore the description

        # we use the safe .get() everywhere below as those elements may not exist
        for array_element in protected_branches_response.get(
            f"{action}_access_levels", []
        ):
            if array_element.get("access_level"):
                levels.add(array_element.get("access_level"))
            elif array_element.get("user_id"):
                user_ids.add(array_element.get("user_id"))
            elif array_element.get("group_id"):
                group_ids.add(array_element.get("group_id"))

        return sorted(levels), sorted(user_ids), sorted(group_ids)

    def get_requested_branch_configuration(self, requested_configuration):
        # from a configuration entry like:
        #
        #   # Allow specific users to push and merge to this branch - *** this syntax requires GitLab Premium (paid) ***
        #   protected: true
        #   allowed_to_push:
        #     - user: jsmith # you can use usernames...
        #     - user: bdoe
        #   allowed_to_merge:
        #     - user_id: 15 # ...or user ids, if you know them
        #
        # ...gets a flat tuple exactly like get_current_branch_protection_settings(), so they can be compared.
        #

        return (
            *(
                self.get_requested_permissions(requested_configuration, "push")
            ),  # tuple of 3
            *(
                self.get_requested_permissions(requested_configuration, "merge")
            ),  # tuple of 3
            *(
                self.get_requested_permissions(requested_configuration, "unprotect")
            ),  # tuple of 3
            requested_configuration.get("allow_force_push", False),
            # code_owner_approval_required has to use PATCH request, see set_code_owner_approval_required()
        )

    def get_requested_permissions(self, requested_configuration, action):
        # from a configuration entry like:
        #
        #   protected: true
        #   allowed_to_push:
        #     - user: jsmith # you can use usernames...
        #     - user: bdoe
        #   allowed_to_merge:
        #     - user_id: 15 # ...or user ids, if you know them
        #
        # ...gets a tuple of sorted lists of entities (levels, users, groups) having the permissions
        # to do the specific action (push/merge/unprotect).
        #
        # for example for "push" action for the config above the return value would be:
        #
        # ([], [123, 456], [])
        #
        # ...where 123, 456 are ids of users jsmith and bdoe

        levels = set()
        user_ids = set()
        group_ids = set()
        # we ignore the description

        if f"{action}_access_level" in requested_configuration:
            levels.add(requested_configuration.get(f"{action}_access_level"))

        if f"allowed_to_{action}" in requested_configuration:
            for config in requested_configuration[f"allowed_to_{action}"]:
                if "access_level" in config:
                    levels.add(config["access_level"])
                elif "user_id" in config:
                    user_ids.add(config["user_id"])
                elif "user" in config:
                    user_ids.add(self.gitlab._get_user_id(config["user"]))
                elif "group_id" in config:
                    group_ids.add(config["group_id"])
                elif "group" in config:
                    group_ids.add(self.gitlab._get_group_id(config["group"]))

        return sorted(levels), sorted(user_ids), sorted(group_ids)



from abc import ABC, abstractmethod
from logging import debug
from typing import Callable, Union

from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.output import EffectiveConfigurationFile
from gitlabform.processors.util.decorators import configuration_to_safe_dict


class AbstractProcessor(ABC):
    def __init__(self, configuration_name: str, gitlab: GitLab):
        self.configuration_name = configuration_name
        self.gitlab = gitlab
        self.custom_diff_analyzers: dict[
            str,
            Callable[
                [str, list[dict[str, Union[str, int]]], list[dict[str, int]]], bool
            ],
        ] = {}

    @configuration_to_safe_dict
    def process(
        self,
        project_or_project_and_group: str,
        configuration,
        dry_run: bool,
        effective_configuration: EffectiveConfigurationFile,
    ):
        if self._section_is_in_config(configuration):
            if configuration.get(f"{self.configuration_name}|skip"):
                verbose(
                    f"Skipping section '{self.configuration_name}' - explicitly configured to do so."
                )
                return
            elif (
                configuration.get("project|archive")
                and self.configuration_name != "project"
            ):
                verbose(
                    f"Skipping section '{self.configuration_name}' - it is configured to be archived."
                )
                return

            if dry_run:
                verbose(
                    f"Processing section '{self.configuration_name}' in dry-run mode."
                )
                self._print_diff(
                    project_or_project_and_group,
                    configuration.get(self.configuration_name),
                )
            else:

                verbose(f"Processing section '{self.configuration_name}'")
                if self._can_proceed(project_or_project_and_group, configuration):
                    self._process_configuration(
                        project_or_project_and_group, configuration
                    )

            effective_configuration.add_configuration(
                project_or_project_and_group,
                self.configuration_name,
                configuration.get(self.configuration_name),
            )
        else:
            verbose(f"Skipping section '{self.configuration_name}' - not in config.")

    def _section_is_in_config(self, configuration: dict):
        return self.configuration_name in configuration

    @abstractmethod
    def _process_configuration(
        self, project_or_project_and_group: str, configuration: dict
    ):
        pass

    def _print_diff(self, project_or_project_and_group: str, entity_config):
        verbose(f"Diffing for section '{self.configuration_name}' is not supported yet")

    def _needs_update(
        self,
        entity_in_gitlab: dict,
        entity_in_configuration: dict,
    ):
        # in the configuration we often don't define every key value because we rely on the defaults.
        # that's why GitLab API often returns many more keys than we have in the configuration.

        # so to decide if the entity should be updated:
        # a) we look for ANY settings that are ONLY in configuration,
        # a) we compare the settings that are in both configuration and gitlab,

        keys_only_in_configuration = set(entity_in_configuration.keys()) - set(
            entity_in_gitlab.keys()
        )
        if len(keys_only_in_configuration) > 0:
            return True

        keys_on_both_sides = set(entity_in_configuration.keys()) & set(
            entity_in_gitlab.keys()
        )
        for key in keys_on_both_sides:
            if key in self.custom_diff_analyzers:
                return self.custom_diff_analyzers[key](
                    key, entity_in_gitlab[key], entity_in_configuration[key]
                )

            if entity_in_gitlab[key] != entity_in_configuration[key]:
                debug(
                    f"entity_in_gitlab[{key}] -> {entity_in_gitlab[key]} != entity_in_configuration[{key}] -> {entity_in_configuration[key]}"
                )
                return True

        return False

    @staticmethod
    def recursive_diff_analyzer(cfg_key: str, cfg_in_gitlab: list, local_cfg: list):
        """
        :return: True if the lists are NOT equal, False otherwise
        """
        if len(cfg_in_gitlab) != len(local_cfg):
            return True

        for index in range(len(cfg_in_gitlab)):
            from_gitlab = {
                k: v for k, v in cfg_in_gitlab[index].items() if v is not None
            }
            from_local_cfg = local_cfg[index]

            keys_on_both_sides = set(from_gitlab.keys()) & set(from_local_cfg.keys())

            for key in keys_on_both_sides:
                if isinstance(from_gitlab[key], list) and isinstance(
                    from_local_cfg[key], list
                ):
                    AbstractProcessor.recursive_diff_analyzer(
                        key, from_gitlab[key], from_local_cfg[key]
                    )

                if from_gitlab[key] != from_local_cfg[key]:
                    debug(
                        f"* A <{key}> in [{cfg_key}] differs:\n GitLab :: {from_gitlab} != Local :: {from_local_cfg}"
                    )
                    return True

        return False

    def _can_proceed(self, project_or_group: str, configuration: dict):
        return True



import base64

from gitlabform.gitlab.core import GitLabCore


class GitLabRepositories(GitLabCore):
    def get_commits_with_string_in_compare_results(
        self, project_and_group_name, c_from, c_to, with_string
    ):
        compare_results = self.compare(project_and_group_name, c_from, c_to)
        commits = compare_results["commits"]
        return [commit for commit in commits if with_string in commit["title"]]

    def compare(self, project_and_group_name, c_from, c_to):
        return self._make_requests_to_api(
            "projects/%s/repository/compare?from=%s&to=%s",
            (project_and_group_name, c_from, c_to),
        )

    def get_file(self, project_and_group_name, branch, path):
        result = self._make_requests_to_api(
            "projects/%s/repository/files/%s?ref=%s",
            (project_and_group_name, path, branch),
        )
        return base64.b64decode(result["content"]).decode("utf-8")

    def set_file(self, project_and_group_name, branch, path, content, commit_message):
        data = {
            "branch": branch,
            "file_path": path,
            "content": content,
            "commit_message": commit_message,
        }
        return self._make_requests_to_api(
            "projects/%s/repository/files/%s",
            (project_and_group_name, path),
            "PUT",
            data=data,
        )

    def add_file(self, project_and_group_name, branch, path, content, commit_message):
        data = {
            "branch": branch,
            "file_path": path,
            "content": content,
            "commit_message": commit_message,
        }
        return self._make_requests_to_api(
            "projects/%s/repository/files/%s",
            (project_and_group_name, path),
            "POST",
            data=data,
            expected_codes=201,
        )

    def delete_file(self, project_and_group_name, branch, path, commit_message):
        data = {"branch": branch, "file_path": path, "commit_message": commit_message}
        return self._make_requests_to_api(
            "projects/%s/repository/files/%s",
            (project_and_group_name, path),
            "DELETE",
            data=data,
            expected_codes=[200, 204],
        )



from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor


class ProjectProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("project", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        if configuration["project"].get("archive"):
            verbose("Archiving project...")
            self.gitlab.archive(project_and_group)
        else:
            verbose("Unarchiving project...")
            self.gitlab.unarchive(project_and_group)



from logging import debug

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor


class HooksProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("hooks", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        for hook in sorted(configuration["hooks"]):

            if configuration.get("hooks|" + hook + "|delete"):
                hook_id = self.gitlab.get_hook_id(project_and_group, hook)
                if hook_id:
                    debug("Deleting hook '%s'", hook)
                    self.gitlab.delete_hook(project_and_group, hook_id)
                else:
                    debug("Not deleting hook '%s', because it doesn't exist", hook)
            else:
                hook_id = self.gitlab.get_hook_id(project_and_group, hook)
                if hook_id:
                    debug("Changing existing hook '%s'", hook)
                    self.gitlab.put_hook(
                        project_and_group, hook_id, hook, configuration["hooks"][hook]
                    )
                else:
                    debug("Creating hook '%s'", hook)
                    self.gitlab.post_hook(
                        project_and_group, hook, configuration["hooks"][hook]
                    )



import logging
import pytest

from gitlabform.configuration import Configuration

logger = logging.getLogger(__name__)


@pytest.fixture
def configuration_for_skip_groups_skip_projects():
    config_yaml = """
    ---
    projects_and_groups:
      "*":
        group_settings:
          one: two
        project_settings:
          three: four

    skip_groups:
    - group-skip
    - group-skip-wildcard/*
    - group-not-skip/subgroup-skip
    - group-not-skip/subgroup-skip-wildcard/*

    skip_projects:
    - project-skip
    - group-skip/project-skip
    - group-skip-wildcard/*
    - group-not-skip/project-skip
    - group-not-skip/subgroup-skip-wildcard/*

    """

    return Configuration(config_string=config_yaml)


@pytest.mark.parametrize(
    "project,is_skipped",
    [
        ("project-skip", True),
        ("project-not-skip", False),
        ("group-not-skip/project-not-skip", False),
        ("group-skip/project-not-skip", False),
        ("group-skip-wildcard/project-skip", True),
        ("group-skip-wildcard/subgroup-a/project-skip", True),
        ("group-skip-wildcard/subgroup-a/subgroup-b/project-skip", True),
        ("group-not-skip/subgroup-not-skip/project-not-skip", False),
        ("group-not-skip/subgroup-skip-wildcard/project-skip", True),
        ("group-not-skip/subgroup-not-skip/project-not-skip", False),
        ("group-not-skip/subgroup-skip/project-not-skip", False),
    ],
)
def test__config_skip_project(project, is_skipped, request):

    configuration_for_skip_groups_skip_projects = request.getfixturevalue(
        "configuration_for_skip_groups_skip_projects"
    )

    assert (
        configuration_for_skip_groups_skip_projects.is_project_skipped(project)
        == is_skipped
    )


@pytest.mark.parametrize(
    "group,is_skipped",
    [
        ("group-not-skip", False),
        ("group-skip", True),
        ("group-skip-wildcard", True),
        ("group-skip-wildcard/subgroup-a", True),
        ("group-skip-wildcard/subgroup-a/subgroup-b", True),
        ("group-not-skip/subgroup-not-skip", False),
        ("group-not-skip/subgroup-skip-wildcard", True),
        ("group-not-skip/subgroup-skip-wildcard/subgroup-skip", True),
        ("group-not-skip/subgroup-not-skip", False),
    ],
)
def test__config_skip_group(group, is_skipped, request):

    configuration_for_skip_groups_skip_projects = request.getfixturevalue(
        "configuration_for_skip_groups_skip_projects"
    )

    assert (
        configuration_for_skip_groups_skip_projects.is_group_skipped(group)
        == is_skipped
    )



import hashlib
import json
from itertools import starmap

from cli_ui import debug as verbose


# Simple function to create strings for values which should be hidden
# example: <secret b2c1a982>
def hide(text: str):
    return f"<secret {hashlib.sha256(text.encode('utf-8')).hexdigest()[:8]}>"


class DifferenceLogger:
    @staticmethod
    def log_diff(
        subject,
        current_config,
        config_to_apply,
        only_changed=False,
        hide_entries=None,
        test=False,
    ):

        # Compose values in list of `[key, from_config, from_server]``
        changes = [
            [
                k,
                json.dumps(
                    current_config.get(k, "???")
                    if type(current_config) == dict
                    else "???"
                ),
                json.dumps(v),
            ]
            for k, v in config_to_apply.items()
        ]

        # Remove unchanged if needed
        if only_changed:
            changes = filter(lambda i: i[1] != i[2], changes)

        # Hide secrets
        if hide_entries:
            changes = list(
                map(
                    lambda i: [i[0], hide(i[1]), hide(i[2])]
                    if i[0] in hide_entries
                    else i,
                    changes,
                )
            )

        # calculate field size for nice formatting
        max_key_len = str(max(map(lambda i: len(i[0]), changes)))
        max_val_1 = str(max(map(lambda i: len(i[1]), changes)))
        max_val_2 = str(max(map(lambda i: len(i[2]), changes)))

        # generate placeholders for output pattern: `     value: before  => after   `
        pattern = (
            "{:>" + max_key_len + "}: {:<" + max_val_1 + "} => {:<" + max_val_2 + "}"
        )

        # create string
        text = "{subject}:\n{diff}".format(
            subject=subject, diff="\n".join(starmap(pattern.format, changes))
        )
        if test:
            return text
        else:
            verbose(text)



from gitlabform.gitlab.core import GitLabCore


class GitLabTags(GitLabCore):
    def get_tags(self, project_and_group_name):
        return self._make_requests_to_api(
            "projects/%s/repository/tags", project_and_group_name
        )

    def create_tag(self, project_and_group_name, tag_name, ref, message=None):
        data = {
            "tag_name": tag_name,
            "ref": ref,
            "message": message,
        }
        return self._make_requests_to_api(
            "projects/%s/repository/tags",
            project_and_group_name,
            method="POST",
            data=data,
            expected_codes=201,
        )

    def delete_tag(self, project_and_group_name, tag_name):
        self._make_requests_to_api(
            "projects/%s/repository/tags/%s",
            (project_and_group_name, tag_name),
            method="DELETE",
            expected_codes=[200, 204],
        )

    def get_protected_tags(self, project_and_group_name):
        return self._make_requests_to_api(
            "projects/%s/protected_tags",
            project_and_group_name,
        )

    def protect_tag(self, project_and_group_name, tag_name, create_access_level):
        data = {"name": tag_name}
        if create_access_level is not None:
            data["create_access_level"] = create_access_level
        return self._make_requests_to_api(
            "projects/%s/protected_tags",
            project_and_group_name,
            method="POST",
            data=data,
            expected_codes=201,
        )

    def unprotect_tag(self, project_and_group_name, tag_name):
        return self._make_requests_to_api(
            "projects/%s/protected_tags/%s",
            (project_and_group_name, tag_name),
            method="DELETE",
            expected_codes=[201, 204],
        )



import pytest

from gitlabform.gitlab.core import NotFoundException
from tests.acceptance import run_gitlabform


@pytest.fixture(scope="function")
def integrations(request, gitlab, group_and_project):

    integrations = ["asana", "slack", "redmine", "jira", "mattermost"]

    def fin():
        # disable test integrations
        for integration in integrations:
            gitlab.delete_integration(group_and_project, integration)

    request.addfinalizer(fin)
    return integrations  # provide fixture value


class TestIntegrations:
    # we use "other_project" here on purpose because if we would reuse the "project"
    # then we could end up with running this test after another, and a integration created
    # and then deleted is a different entity in GitLab than a never created one (!).
    # the first one exists but has "active" field set to False, the other throws 404
    def test__if_they_are_not_set_by_default(self, gitlab, group, other_project):
        group_and_project = f"{group}/{other_project}"

        for integration_name in ["asana", "slack", "redmine", "jira"]:
            with pytest.raises(NotFoundException):
                gitlab.get_integration(group_and_project, integration_name)

    def test__if_delete_works(self, gitlab, group_and_project):

        config_integrations = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              jira:
                url: http://foo.bar.com
                username: foo
                password: bar
                active: true
                commit_events: true
              asana:
                api_key: foo
                push_events: true
              slack:
                webhook: http://foo.bar.com
                push_events: true
              redmine:
                new_issue_url: http://foo.bar.com
                project_url: http://foo.bar.com
                issues_url: http://foo.bar.com
                push_events: true
        """

        run_gitlabform(config_integrations, group_and_project)

        for integration_name in ["jira", "asana", "slack", "redmine"]:
            integration = gitlab.get_integration(group_and_project, integration_name)
            assert integration["active"] is True

        config_integrations_delete = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              jira:
                delete: true
              slack:
                delete: true
        """

        run_gitlabform(config_integrations_delete, group_and_project)

        for integration_name in ["jira", "slack"]:
            integration = gitlab.get_integration(group_and_project, integration_name)
            assert integration["active"] is False

    def test__if_push_events_true_works(self, gitlab, group_and_project):

        config_integration_push_events_true = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              asana:
                api_key: foo
                push_events: true
              slack:
                webhook: http://foo.bar.com
                push_events: true
              redmine:
                new_issue_url: http://foo.bar.com
                project_url: http://foo.bar.com
                issues_url: http://foo.bar.com
                push_events: true
        """

        run_gitlabform(config_integration_push_events_true, group_and_project)

        integrations = []
        for integration_name in ["asana", "slack", "redmine"]:
            integration = gitlab.get_integration(group_and_project, integration_name)
            integrations.append(integration)

        assert all([integration["active"] for integration in integrations]) is True
        assert all([integration["push_events"] for integration in integrations]) is True

    def test__if_push_events_false_works(self, gitlab, group_and_project):

        config_integration_push_events_false = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              asana:
                api_key: foo
                push_events: false # changed
              slack:
                webhook: http://foo.bar.com
                push_events: false # changed
              redmine:
                new_issue_url: http://foo.bar.com
                project_url: http://foo.bar.com
                issues_url: http://foo.bar.com
                push_events: false # changed
        """

        run_gitlabform(config_integration_push_events_false, group_and_project)

        integrations = []
        for integration_name in ["asana", "slack", "redmine"]:
            integration = gitlab.get_integration(group_and_project, integration_name)
            integrations.append(integration)

        assert all([integration["active"] for integration in integrations]) is True
        assert (
            all([integration["push_events"] for integration in integrations]) is False
        )

    def test__if_push_events_change_works(self, gitlab, group_and_project):

        config_integration_push_events_true = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              asana:
                api_key: foo
                push_events: true
              slack:
                webhook: http://foo.bar.com
                push_events: true
              redmine:
                new_issue_url: http://foo.bar.com
                project_url: http://foo.bar.com
                issues_url: http://foo.bar.com
                push_events: true
        """

        run_gitlabform(config_integration_push_events_true, group_and_project)

        integrations = []
        for integration_name in ["asana", "slack", "redmine"]:
            integration = gitlab.get_integration(group_and_project, integration_name)
            integrations.append(integration)

        assert all([integration["active"] for integration in integrations]) is True
        assert all([integration["push_events"] for integration in integrations]) is True

        config_integration_push_events_false = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              asana:
                api_key: foo
                push_events: false # changed
              slack:
                webhook: http://foo.bar.com
                push_events: false # changed
              redmine:
                new_issue_url: http://foo.bar.com
                project_url: http://foo.bar.com
                issues_url: http://foo.bar.com
                push_events: false # changed
        """

        run_gitlabform(config_integration_push_events_false, group_and_project)

        integrations = []
        for integration_name in ["asana", "slack", "redmine"]:
            integration = gitlab.get_integration(group_and_project, integration_name)
            integrations.append(integration)

        assert all([integration["active"] for integration in integrations]) is True
        assert (
            all([integration["push_events"] for integration in integrations]) is False
        )

    def test__if_jira_commit_events_true_works(self, gitlab, group_and_project):

        config_integration_jira_commit_events_true = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              jira:
                url: http://foo.bar.com
                username: foo
                password: bar
                active: true
                commit_events: true
        """

        run_gitlabform(config_integration_jira_commit_events_true, group_and_project)

        integration = gitlab.get_integration(group_and_project, "jira")
        assert integration["active"] is True
        assert integration["commit_events"] is True

    def test__if_jira_commit_events_false_works(self, gitlab, group_and_project):

        config_integration_jira_commit_events_false = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              jira:
                url: http://foo.bar.com
                username: foo
                password: bar
                active: true
                commit_events: false
        """

        run_gitlabform(config_integration_jira_commit_events_false, group_and_project)

        integration = gitlab.get_integration(group_and_project, "jira")
        assert integration["active"] is True
        assert integration["commit_events"] is False

    def test__if_change_works(self, gitlab, group_and_project):

        config_integration_jira_commit_events_true = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              jira:
                url: http://foo.bar.com
                username: foo
                password: bar
                active: true
                commit_events: true
        """

        run_gitlabform(config_integration_jira_commit_events_true, group_and_project)

        integration = gitlab.get_integration(group_and_project, "jira")
        assert integration["active"] is True
        assert integration["commit_events"] is True

        config_integration_jira_commit_events_false = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              jira:
                url: http://foo.bar.com
                username: foo
                password: bar
                active: true
                commit_events: false
        """

        run_gitlabform(config_integration_jira_commit_events_false, group_and_project)

        integration = gitlab.get_integration(group_and_project, "jira")
        assert integration["active"] is True
        assert integration["commit_events"] is False

    def test__mattermost_confidential_issues_events(self, gitlab, group_and_project):

        config_integration_mattermost_confidential_issues_events = f"""
        projects_and_groups:
          {group_and_project}:
            integrations:
              mattermost:
                active: true
                webhook: https://mattermost.com/hooks/xxx
                username: gitlab
                merge_requests_events: true
                merge_request_channel: "merge-requests"
                push_events: false
                issues_events: false
                confidential_issues_events: false # this was not supposed to work according to #70
                tag_push_events: false
                note_events: false
                confidential_note_events: false
                pipeline_events: false
                wiki_page_events: false
                branches_to_be_notified: "all"
        """

        run_gitlabform(
            config_integration_mattermost_confidential_issues_events, group_and_project
        )

        integration = gitlab.get_integration(group_and_project, "mattermost")
        assert integration["confidential_issues_events"] is False



import logging

from gitlabform.configuration import Configuration

logger = logging.getLogger(__name__)


class TestInheritanceBreakProjectsAndGroups:
    def test__inheritance_break__flag_set_at_project_level__project_inherits_nothing(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          "*":
            variables:
              secret1:
                key: foo
                value: bar

          "some_group/*":
            variables:
              secret2:
                key: foo
                value: bar

          "some_group/my_project":
            variables:
              inherit: false
              secret3:
                key: bizz
                value: buzz
        """

        configuration = Configuration(config_string=config_yaml)

        effective_config = configuration.get_effective_config_for_project(
            "some_group/my_project"
        )

        variables = effective_config["variables"]

        assert variables == {
            "secret3": {"key": "bizz", "value": "buzz"},
        }

    def test__inheritance_break__flag_set_at_project_level__project_inherits_group_and_not_common(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          "*":
            variables:
              secret1:
                key: foo
                value: bar

          "some_group/*":
            members:
              users:
                user1: developer
                user2: developer

          "some_group/my_project":
            variables:
              inherit: false
              secret2:
                key: bizz
                value: buzz
        """

        configuration = Configuration(config_string=config_yaml)

        effective_config = configuration.get_effective_config_for_project(
            "some_group/my_project"
        )

        assert effective_config == {
            "members": {
                "users": {
                    "user1": "developer",
                    "user2": "developer",
                }
            },
            "variables": {
                "secret2": {"key": "bizz", "value": "buzz"},
            },
        }

    def test__inheritance_break__flag_set_at_group_level__group_inherits_nothing(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          "*":
            variables:
              secret1:
                key: foo
                value: bar

          "some_group/*":
            variables:
              inherit: false
              secret2:
                key: foo
                value: bar
        """

        configuration = Configuration(config_string=config_yaml)

        effective_config = configuration.get_effective_config_for_group("some_group")

        variables = effective_config["variables"]

        assert variables == {
            "secret2": {"key": "foo", "value": "bar"},
        }

    def test__inheritance_break__flag_set_at_group_level__project_inherits_group_and_not_common(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          "*":
            variables:
              secret1:
                key: foo
                value: bar

          "some_group/*":
            variables:
              inherit: false
              secret2:
                key: foo
                value: bar

          "some_group/my_project":
            variables:
              secret3:
                key: bizz
                value: buzz
        """

        configuration = Configuration(config_string=config_yaml)

        effective_config = configuration.get_effective_config_for_project(
            "some_group/my_project"
        )

        variables = effective_config["variables"]

        assert variables == {
            "secret2": {"key": "foo", "value": "bar"},
            "secret3": {"key": "bizz", "value": "buzz"},
        }

    def test__inheritance_break__flag_set_at_group_level_and_project_level__project_inherits_nothing(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          "*":
            variables:
              secret1:
                key: foo
                value: bar

          "some_group/*":
            variables:
              inherit: false
              secret2:
                key: foo
                value: bar

          "some_group/my_project":
            variables:
              inherit: false
              secret3:
                key: bizz
                value: buzz
        """

        configuration = Configuration(config_string=config_yaml)

        effective_config = configuration.get_effective_config_for_project(
            "some_group/my_project"
        )

        variables = effective_config["variables"]

        assert variables == {
            "secret3": {"key": "bizz", "value": "buzz"},
        }



from gitlabform.gitlab.projects import GitLabProjects


class GitLabVariables(GitLabProjects):
    def get_variables(self, project_and_group_name):
        return self._make_requests_to_api(
            "projects/%s/variables", project_and_group_name
        )

    def post_variable(self, project_and_group_name, variable_in_config):
        # variable has to be like documented at:
        # https://docs.gitlab.com/ee/api/project_level_variables.html#create-variable
        self._make_requests_to_api(
            "projects/%s/variables",
            project_and_group_name,
            "POST",
            variable_in_config,
            expected_codes=201,
        )

    def put_variable(
        self,
        project_and_group_name,
        variable_in_gitlab,
        variable_in_config,
    ):
        # variable has to be like documented at:
        # https://docs.gitlab.com/ce/api/build_variables.html#update-variable
        self._make_requests_to_api(
            "projects/%s/variables/%s",
            (project_and_group_name, variable_in_gitlab["key"]),
            "PUT",
            variable_in_config,
        )

    def delete_variable(self, project_and_group_name, variable_in_config):
        self._make_requests_to_api(
            "projects/%s/variables/%s",
            (project_and_group_name, variable_in_config["key"]),
            method="DELETE",
            expected_codes=[204, 404],
        )

    def get_variable(self, project_and_group_name, variable_key, environment_scope="*"):
        if environment_scope == "*":
            url = "projects/%s/variables/%s"
        else:
            url = f"projects/%s/variables/%s?filter[environment_scope]={environment_scope}"
        return self._make_requests_to_api(url, (project_and_group_name, variable_key))[
            "value"
        ]



import logging
import pytest

from gitlabform.configuration import Configuration

logger = logging.getLogger(__name__)


@pytest.fixture
def configuration_with_only_group_and_project():
    config_yaml = """
    ---
    projects_and_groups:
      some_group/*:
        project_settings:
          foo: bar
        hooks:
          a:
            foo: bar

      some_group/some_project:
        project_settings:
          bar: foo
        hooks:
          b:
            bar: foo
    """

    return Configuration(config_string=config_yaml)


def test__get_effective_config_for_project__other_project(
    configuration_with_only_group_and_project,
):
    effective_config = (
        configuration_with_only_group_and_project.get_effective_config_for_project(
            "project_not_in_config/group_not_in_config"
        )
    )

    assert effective_config == {}


def test__get_effective_config_for_project__project_from_config__additive_project_settings(
    configuration_with_only_group_and_project,
):
    effective_config = (
        configuration_with_only_group_and_project.get_effective_config_for_project(
            "some_group/some_project"
        )
    )

    additive__project_settings = effective_config["project_settings"]

    # merged hashes from group and project levels
    assert additive__project_settings == {"foo": "bar", "bar": "foo"}


def test__get_effective_config_for_project__project_from_config__additive_hooks(
    configuration_with_only_group_and_project,
):
    effective_config = (
        configuration_with_only_group_and_project.get_effective_config_for_project(
            "some_group/some_project"
        )
    )

    additive__hooks = effective_config["hooks"]
    assert additive__hooks == {
        "a": {"foo": "bar"},
        "b": {"bar": "foo"},
    }  # added from both group and project level


def test__get_effective_config_for_project__with_multiple_levels():
    config_yaml = """
    ---
    projects_and_groups:
      "*":
        merge_requests:
          approvals:
            approvals_before_merge: 3
          approvers:
            - common_approvers

      "some_group/*":
        merge_requests:
          approvals:
            reset_approvals_on_push: true
          approvers:
            - group_approvers

      "some_group/my_project":
        merge_requests:
          approvals:
            reset_approvals_on_push: false
            disable_overriding_approvers_per_merge_request: true
          approvers:
            - project_approvers
    """

    effective_config = Configuration(
        config_string=config_yaml
    ).get_effective_config_for_project("some_group/my_project")

    assert effective_config == {
        "merge_requests": {
            "approvals": {
                # dict is merged - below is the config from common + group + project level
                "approvals_before_merge": 3,
                "reset_approvals_on_push": False,
                "disable_overriding_approvers_per_merge_request": True,
            },
            "approvers": [
                # list is overriden - only the element from the project level config is here
                "project_approvers",
            ],
        }
    }


def test__get_effective_config_for_project__configuration_for_other_project():
    config_yaml = """
    ---
    projects_and_groups:
      "some_group/*":
        variables:
          first:
            key: foo
            value: bar

      "some_group/my_project":
        variables:
          second:
            key: foo
            value: bar
    """

    effective_config = Configuration(
        config_string=config_yaml
    ).get_effective_config_for_project("some_group/some_project_after_defined")

    assert effective_config == {
        "variables": {
            "first": {
                "key": "foo",
                "value": "bar",
            },
        }
    }



from cli_ui import fatal

from gitlabform.constants import EXIT_INVALID_INPUT
from gitlabform.lists import OmissionReason, Groups


from gitlabform.gitlab.core import NotFoundException


class GroupsProvider:
    """
    For a query like "project/group", "group", "group/subgroup", ALL or ALL_DEFINED this
    class gets the effective lists of groups, taking into account skipped groups
    and the fact that the group and project names case are somewhat case-sensitive.
    """

    def __init__(self, gitlab, configuration):
        self.gitlab = gitlab
        self.configuration = configuration

    def get_groups(self, target: str) -> Groups:
        """
        :param target: "project/group", "group", "group/subgroup", ALL or ALL_DEFINED
        :return: Groups
        """

        if target not in ["ALL", "ALL_DEFINED"]:
            groups = self._get_single_group(target)
        else:
            groups = self._get_groups(target)

        return groups

    def _get_single_group(self, target: str) -> Groups:

        groups = Groups()

        # it may be a subgroup or a group...
        try:
            maybe_group = self.gitlab.get_group_case_insensitive(target)
            groups.add_requested([maybe_group["full_path"]])

        except NotFoundException:
            # ...or a single project, which we ignore here
            pass

        return groups

    def _get_groups(self, target: str) -> Groups:

        groups = Groups()

        if target == "ALL":
            groups.add_requested(self.gitlab.get_groups())
        else:  # ALL_DEFINED
            groups.add_requested(self.configuration.get_groups())

        groups.add_omitted(
            OmissionReason.SKIPPED, self._get_skipped_groups(groups.get_effective())
        )

        if target == "ALL_DEFINED":
            # check if all the groups from the config actually exist
            self._verify_if_groups_exist(groups.get_effective())

        return groups

    def _verify_if_groups_exist(self, groups: list):
        for group in groups:
            try:
                self.gitlab.get_group_case_insensitive(group)
            except NotFoundException:
                fatal(
                    f"Configuration contains group {group} but it cannot be found in GitLab!",
                    exit_code=EXIT_INVALID_INPUT,
                )

    def _get_skipped_groups(self, groups: list) -> list:
        skipped = []
        for group in groups:
            if self.configuration.is_group_skipped(group):
                skipped.append(group)

        return skipped



from abc import ABC, abstractmethod


class AbstractKey(ABC):
    """
    This represents a key in an entity configuration.

    For example in such an entity:

    rule:
      name: foo
      rule_type: bar

    ...the "name" and "rule_type" are keys.

    This may also be an expression made with keys and relationships between them, see below.
    """

    @abstractmethod
    def matches(self, e1, e2):
        """
        :param e1: some entity
        :param e2: another entity
        :return: True if the two given entities have the same value of this key.
        """
        pass

    @abstractmethod
    def contains(self, entity):
        """
        :param entity: some entity
        :return: True if the entity contains this key
        """
        pass

    @abstractmethod
    def explain(self) -> str:
        """
        :return: a user-friendly explanation of what this key is
        """
        pass


class Key(AbstractKey):
    """
    A single, mandatory key.
    """

    def __init__(self, name):
        self.name = name

    def matches(self, e1, e2):
        return self.name in e1 and self.name in e2 and e1[self.name] == e2[self.name]

    def contains(self, entity):
        return entity.get(self.name, None) is not None

    def explain(self) -> str:
        return f"'{self.name}'"


class And(AbstractKey):
    """
    This groups two or more keys of which all are mandatory.
    """

    def __init__(self, *arg: AbstractKey):
        self.keys = arg

    def matches(self, e1, e2):
        return all([key.matches(e1, e2) for key in self.keys])

    def contains(self, entity):
        return all([key.contains(entity) for key in self.keys])

    def explain(self) -> str:
        explains = [key.explain() for key in self.keys]
        return f"({' and '.join(explains)})"


class Or(AbstractKey):
    """
    This groups two or more keys where at least one must exist.
    """

    def __init__(self, *arg: AbstractKey):
        self.keys = arg

    def matches(self, e1, e2):
        return any([key.matches(e1, e2) for key in self.keys])

    def contains(self, entity):
        return any([key.contains(entity) for key in self.keys])

    def explain(self) -> str:
        explains = [key.explain() for key in self.keys]
        return f"({' or '.join(explains)})"


class Xor(AbstractKey):
    """
    This groups two or more keys where exactly one must exist.
    """

    def __init__(self, *arg: AbstractKey):
        self.keys = arg

    # copied from https://stackoverflow.com/a/16801336/2693875
    @staticmethod
    def _single_true(iterable):
        iterator = iter(iterable)

        # consume from "i" until first true, or it's exhausted
        has_true = any(iterator)

        # carry on consuming until another true value / exhausted
        has_another_true = any(iterator)

        # True if exactly one true found
        return has_true and not has_another_true

    def matches(self, e1, e2):
        return self._single_true([key.matches(e1, e2) for key in self.keys])

    def contains(self, entity):
        return self._single_true([key.contains(entity) for key in self.keys])

    def explain(self) -> str:
        explains = [key.explain() for key in self.keys]
        return f"(exactly one of: {', '.join(explains)})"


class OptionalKey(AbstractKey):
    """
    This is a non-mandatory key.
    """

    def __init__(self, name):
        self.name = name

    def matches(self, e1, e2):
        only_in_e1 = self.name in e1 and self.name not in e2
        only_in_e2 = self.name not in e1 and self.name in e2
        in_both_and_equal = (
            self.name in e1 and self.name in e2 and e1[self.name] == e2[self.name]
        )
        return only_in_e1 or only_in_e2 or in_both_and_equal

    def contains(self, entity):
        # as it's optional, we don't check for its existence
        return True

    def explain(self) -> str:
        return f"(optionally '{self.name}')"



from typing import List

from gitlabform.configuration import Configuration
from gitlabform.gitlab import GitLab
from gitlabform.processors import AbstractProcessors
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.group.group_badges_processor import GroupBadgesProcessor
from gitlabform.processors.group.group_ldap_links_processor import (
    GroupLDAPLinksProcessor,
)
from gitlabform.processors.group.group_members_processor import (
    GroupMembersProcessor,
)
from gitlabform.processors.group.group_variables_processor import (
    GroupVariablesProcessor,
)
from gitlabform.processors.group.group_settings_processor import (
    GroupSettingsProcessor,
)


class GroupProcessors(AbstractProcessors):
    def __init__(self, gitlab: GitLab, config: Configuration, strict: bool):
        super().__init__(gitlab, config, strict)
        self.processors: List[AbstractProcessor] = [
            GroupVariablesProcessor(gitlab),
            GroupSettingsProcessor(gitlab),
            GroupMembersProcessor(gitlab),
            GroupLDAPLinksProcessor(gitlab),
            GroupBadgesProcessor(gitlab),
        ]



from gitlabform.gitlab.core import UnexpectedResponseException
from gitlabform.gitlab.projects import GitLabProjects


class GitLabProjectDeployKeys(GitLabProjects):
    def get_all_deploy_keys(self):
        return self._make_requests_to_api("deploy_keys")

    def get_deploy_keys(self, project_and_group_name):
        return self._make_requests_to_api(
            "projects/%s/deploy_keys", project_and_group_name
        )

    def post_deploy_key(self, project_and_group_name, deploy_key_in_config):

        # deploy_key_in_config has to be like this:
        # {
        #     'title': title,
        #     'key': key,
        #     'can_push': can_push,
        # }
        # ..as documented at: https://docs.gitlab.com/ce/api/deploy_keys.html#add-deploy-key

        try:
            self._make_requests_to_api(
                "projects/%s/deploy_keys",
                project_and_group_name,
                "POST",
                deploy_key_in_config,
                expected_codes=[201],
            )

        except UnexpectedResponseException as e:
            if e.response_status_code == 400 and (
                "has already been taken" in e.response_text
            ):

                # Sometimes GitLab throws HTTP 400: {"deploy_key.fingerprint_sha256":["has already been taken"]}
                # when you try to add an existing SSH key to another project, although according to the API docs
                # it should work
                # (see https://docs.gitlab.com/ee/api/deploy_keys.html#add-deploy-keys-to-multiple-projects).

                # As a workaround, when that happens, we will try to get this key's id and enable this existing key
                # for the new project.

                all_existing_keys = self.get_all_deploy_keys()

                existing_key_id = None
                for existing_key in all_existing_keys:
                    if self._keys_are_effectively_equal(
                        existing_key["key"], deploy_key_in_config["key"]
                    ):
                        existing_key_id = existing_key["id"]
                        break

                if existing_key_id:
                    # POST /projects/:id/deploy_keys/:key_id/enable
                    self._make_requests_to_api(
                        "projects/%s/deploy_keys/%s/enable",
                        (project_and_group_name, existing_key_id),
                        "POST",
                        expected_codes=201,
                    )
                else:
                    e.message = (
                        e.message
                        + "\n"
                        + "!!! However we haven't found this key in GitLab under the /deploy_keys API... !!!"
                    )
                    raise e

    def put_deploy_key(
        self, project_and_group_name, deploy_key_in_gitlab, deploy_key_in_config
    ):
        # according to docs_new at https://docs.gitlab.com/ee/api/deploy_keys.html#update-deploy-key
        # you only can change key's title and can_push param

        changeable_data = {
            "title": deploy_key_in_config.get("title", None),
            "can_push": deploy_key_in_config.get("can_push", None),
        }

        return self._make_requests_to_api(
            "projects/%s/deploy_keys/%s",
            (project_and_group_name, deploy_key_in_gitlab["id"]),
            "PUT",
            changeable_data,
        )

    def delete_deploy_key(self, project_and_group_name, deploy_key_in_config):
        return self._make_requests_to_api(
            "projects/%s/deploy_keys/%s",
            (project_and_group_name, deploy_key_in_config["id"]),
            method="DELETE",
            expected_codes=[204, 404],
        )

    def get_deploy_key(self, project_and_group_name, id):
        return self._make_requests_to_api(
            "projects/%s/deploy_keys/%s", (project_and_group_name, id), "GET"
        )

    @staticmethod
    def _keys_are_effectively_equal(key1, key2):
        # We ignore the comment part of the SSH key as GitLab doesn't allow adding the same key just
        # with a different comment BUT it also has a bug that it returns keys with only parts of the
        # comments if the comment contains spaces, so it may show a difference where there is none...

        key1_type = key1.split(" ")[0]
        key1_value = key1.split(" ")[1]
        key2_type = key2.split(" ")[0]
        key2_value = key2.split(" ")[1]

        return (key1_type == key2_type) and (key1_value == key2_value)



from logging import debug
from typing import Dict, List

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor


class SchedulesProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("schedules", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        existing_schedules = self.gitlab.get_all_pipeline_schedules(project_and_group)
        schedule_ids_by_description = self.__group_schedule_ids_by_description(
            existing_schedules
        )

        for schedule_description in sorted(configuration["schedules"]):
            schedule_ids = schedule_ids_by_description.get(schedule_description)
            if configuration.get("schedules|" + schedule_description + "|delete"):
                if schedule_ids:
                    debug("Deleting pipeline schedules '%s'", schedule_description)
                    for schedule_id in schedule_ids:
                        self.gitlab.delete_pipeline_schedule(
                            project_and_group, schedule_id
                        )
                else:
                    debug(
                        "Not deleting pipeline schedules '%s', because none exist",
                        schedule_description,
                    )
            else:
                if schedule_ids and len(schedule_ids) == 1:
                    debug(
                        "Changing existing pipeline schedule '%s'", schedule_description
                    )
                    schedule_id = schedule_ids[0]
                    self.gitlab.take_ownership(project_and_group, schedule_id)
                    self.gitlab.update_pipeline_schedule(
                        project_and_group,
                        schedule_id,
                        configuration.get("schedules|" + schedule_description),
                    )
                    self.__set_schedule_variables(
                        project_and_group,
                        schedule_id,
                        configuration.get(
                            "schedules|" + schedule_description + "|variables"
                        ),
                    )
                elif schedule_ids:
                    debug(
                        "Replacing existing pipeline schedules '%s'",
                        schedule_description,
                    )
                    for schedule_id in schedule_ids:
                        self.gitlab.delete_pipeline_schedule(
                            project_and_group, schedule_id
                        )
                    self.create_schedule_with_variables(
                        configuration, project_and_group, schedule_description
                    )
                else:
                    debug("Creating pipeline schedule '%s'", schedule_description)
                    self.create_schedule_with_variables(
                        configuration, project_and_group, schedule_description
                    )

    def create_schedule_with_variables(
        self, configuration, project_and_group, schedule_description
    ):
        data = configuration.get("schedules|" + schedule_description)
        created_schedule = self.gitlab.create_pipeline_schedule(
            project_and_group,
            schedule_description,
            data.get("ref"),
            data.get("cron"),
            optional_data=data,
        )
        self.__set_schedule_variables(
            project_and_group,
            created_schedule.get("id"),
            configuration.get("schedules|" + schedule_description + "|variables"),
        )

    def __set_schedule_variables(self, project_and_group, schedule_id, variables):
        schedule = self.gitlab.get_pipeline_schedule(project_and_group, schedule_id)

        existing_variables = schedule.get("variables")
        if existing_variables:
            debug(
                "Deleting variables for pipeline schedule '%s'", schedule["description"]
            )

            for variable in existing_variables:
                self.gitlab.delete_pipeline_schedule_variable(
                    project_and_group, schedule_id, variable.get("key")
                )

        if variables:
            for variable_key, variable_data in variables.items():
                self.gitlab.create_pipeline_schedule_variable(
                    project_and_group,
                    schedule_id,
                    variable_key,
                    variable_data.get("value"),
                    variable_data,
                )

    @staticmethod
    def __group_schedule_ids_by_description(schedules) -> Dict[str, List[str]]:
        schedule_ids_by_description: Dict[str, List[str]] = {}

        for schedule in schedules:
            description = schedule["description"]
            schedule_ids_by_description.setdefault(description, []).append(
                schedule["id"]
            )

        return schedule_ids_by_description



from gitlabform.gitlab import GitLab
from gitlabform.processors.defining_keys import And, Key
from gitlabform.processors.multiple_entities_processor import MultipleEntitiesProcessor


class DeployKeysProcessor(MultipleEntitiesProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "deploy_keys",
            gitlab,
            list_method_name="get_deploy_keys",
            add_method_name="post_deploy_key",
            delete_method_name="delete_deploy_key",
            defining=Key("title"),
            required_to_create_or_update=And(Key("title"), Key("key")),
            # DO NOT use put_deploy_key for update as it can only update key's title,
            # but NOT the value (according to https://docs.gitlab.com/ee/api/deploy_keys.html#update-deploy-key)
        )



import logging
import pytest

from gitlabform.configuration import Configuration

logger = logging.getLogger(__name__)


class TestInheritanceBreakSubgroups:
    def test__inheritance_break__flag_set_at_subgroup_level__subgroup_inherits_nothing(
        self,
    ):
        config_yaml = """
        ---
        projects_and_groups:
          some_group/*:
            project_settings:
              foo: bar
    
          some_group/some_subgroup/*:
            project_settings:
              inherit: false
              fizz: buzz
        """

        configuration = Configuration(config_string=config_yaml)

        effective_config = configuration.get_effective_config_for_group(
            "some_group/some_subgroup"
        )

        assert effective_config == {
            "project_settings": {"fizz": "buzz"},
        }

    @pytest.fixture
    def configuration_with_inheritance_break_set_at_subgroup_and_project_level(self):
        config_yaml = """
        ---
        projects_and_groups:
          some_group/*:
            group_settings:
              snap: crackle
            project_settings:
              foo: bar

          some_group/subgroup_level_1/*:
            group_settings:
              inherit: false
              crackle: pop
            project_settings:
              foo1: bar1
              
          some_group/subgroup_level_1/some_project:
            project_settings:
              inherit: false
              fizz: buzz

          some_group/subgroup_level_1/subgroup_level_2/*:
            project_settings:
              inherit: false
              foo2: bar2

          some_group/subgroup_level_1/subgroup_level_2/some_project:
            group_settings:
              inherit: false
              snap: pop
            project_settings:
              inherit: false
              fizz: buzz
        """
        return Configuration(config_string=config_yaml)

    def test__inheritance_break__flag_set_at_subgroup_level__first_subgroup_does_not_inherit_group_settings(
        self,
        configuration_with_inheritance_break_set_at_subgroup_and_project_level,
    ):
        effective_config = configuration_with_inheritance_break_set_at_subgroup_and_project_level.get_effective_config_for_group(
            "some_group/subgroup_level_1"
        )

        assert effective_config == {
            "group_settings": {"crackle": "pop"},
            "project_settings": {
                "foo": "bar",
                "foo1": "bar1",
            },
        }

    def test__inheritance_break__flag_set_at_project_level__first_subgroup_project_does_not_inherit_project_settings(
        self,
        configuration_with_inheritance_break_set_at_subgroup_and_project_level,
    ):
        effective_config = configuration_with_inheritance_break_set_at_subgroup_and_project_level.get_effective_config_for_project(
            "some_group/subgroup_level_1/some_project"
        )

        assert effective_config == {
            "group_settings": {"crackle": "pop"},
            "project_settings": {
                "fizz": "buzz",
            },
        }

    def test__inheritance_break__flag_set_at_subgroup_level__second_subgroup_does_not_inherit_project_settings(
        self,
        configuration_with_inheritance_break_set_at_subgroup_and_project_level,
    ):
        effective_config = configuration_with_inheritance_break_set_at_subgroup_and_project_level.get_effective_config_for_group(
            "some_group/subgroup_level_1/subgroup_level_2"
        )

        assert effective_config == {
            "group_settings": {"crackle": "pop"},
            "project_settings": {
                "foo2": "bar2",
            },
        }

    def test__inheritance_break__flag_set_at_project_level__second_subgroup_project_inherits_nothing(
        self,
        configuration_with_inheritance_break_set_at_subgroup_and_project_level,
    ):
        effective_config = configuration_with_inheritance_break_set_at_subgroup_and_project_level.get_effective_config_for_project(
            "some_group/subgroup_level_1/subgroup_level_2/some_project"
        )

        assert effective_config == {
            "group_settings": {"snap": "pop"},
            "project_settings": {
                "fizz": "buzz",
            },
        }



import pytest


from tests.acceptance import run_gitlabform


@pytest.fixture(scope="class")
def schedules(gitlab, group_and_project):

    another_branch = "scheduled/new-feature"
    gitlab.create_branch(group_and_project, another_branch, "main")

    # fmt: off
    schedules = [
        ("Existing schedule", "main", "0 * * * *", {"active": "true"}),
        ("Existing schedule with vars", "main", "30 * * * *", {}),
        ("Existing schedule to replace", "main", "30 1 * * *", {}),
        ("Existing schedule to replace", another_branch, "30 2 * * *", {}),
    ]
    # fmt: on

    for schedule in schedules:
        gitlab.create_pipeline_schedule(
            group_and_project,
            *schedule,
        )

    redundant_schedule = gitlab.create_pipeline_schedule(
        group_and_project,
        "Redundant schedule",
        "main",
        "0 * * * *",
    )
    gitlab.create_pipeline_schedule_variable(
        group_and_project, redundant_schedule["id"], "test_variable", "test_value"
    )

    schedules.append(("Redundant schedule", "main", "0 * * * *"))

    return schedules  # provide fixture value


class TestSchedules:
    def test__add_new_schedule(self, gitlab, group_and_project, schedules):

        add_schedule = f"""
        projects_and_groups:
          {group_and_project}:
            schedules:
              "New schedule":
                ref: main
                cron: "0 * * * *"
                cron_timezone: "London"
                active: true
        """

        run_gitlabform(add_schedule, group_and_project)

        schedule = self.__find_pipeline_schedule_by_description_and_get_first(
            gitlab, group_and_project, "New schedule"
        )
        assert schedule is not None
        assert schedule["description"] == "New schedule"
        assert schedule["ref"] == "main"
        assert schedule["cron"] == "0 * * * *"
        assert schedule["cron_timezone"] == "London"
        assert schedule["active"] is True

    def test__add_new_schedule_with_mandatory_fields_only(
        self, gitlab, group_and_project, schedules
    ):

        add_schedule_mandatory_fields_only = f"""
        projects_and_groups:
          {group_and_project}:
            schedules:
              "New schedule with mandatory fields":
                ref: main
                cron: "30 1 * * *"
        """

        run_gitlabform(add_schedule_mandatory_fields_only, group_and_project)

        schedule = self.__find_pipeline_schedule_by_description_and_get_first(
            gitlab, group_and_project, "New schedule with mandatory fields"
        )
        assert schedule is not None
        assert schedule["description"] == "New schedule with mandatory fields"
        assert schedule["ref"] == "main"
        assert schedule["cron"] == "30 1 * * *"
        assert schedule["cron_timezone"] == "UTC"
        assert schedule["active"] is True

    def test__set_schedule_variables(self, gitlab, group_and_project, schedules):

        add_schedule_with_variables = f"""
        projects_and_groups:
          {group_and_project}:
            schedules:
              "New schedule with variables":
                ref: main
                cron: "30 1 * * *"
                variables:
                    var1:
                        value: value123
                    var2:
                        value: value987
                        variable_type: file
        """

        run_gitlabform(add_schedule_with_variables, group_and_project)

        schedule = self.__find_pipeline_schedule_by_description_and_get_first(
            gitlab, group_and_project, "New schedule with variables"
        )
        assert schedule is not None
        assert schedule["description"] == "New schedule with variables"
        assert schedule["ref"] == "main"
        assert schedule["cron"] == "30 1 * * *"
        assert schedule["cron_timezone"] == "UTC"
        assert schedule["active"] is True
        assert schedule["variables"] is not None
        assert len(schedule["variables"]) == 2

        assert schedule["variables"][0]["variable_type"] == "env_var"
        assert schedule["variables"][0]["key"] == "var1"
        assert schedule["variables"][0]["value"] == "value123"

        assert schedule["variables"][1]["variable_type"] == "file"
        assert schedule["variables"][1]["key"] == "var2"
        assert schedule["variables"][1]["value"] == "value987"

    def test__update_existing_schedule(self, gitlab, group_and_project, schedules):

        edit_schedule = f"""
        projects_and_groups:
          {group_and_project}:
            schedules:
              "Existing schedule":
                ref: scheduled/new-feature
                cron: "0 */4 * * *"
                cron_timezone: "Stockholm"
                active: false
        """

        run_gitlabform(edit_schedule, group_and_project)

        schedule = self.__find_pipeline_schedule_by_description_and_get_first(
            gitlab, group_and_project, "Existing schedule"
        )
        assert schedule is not None
        assert schedule["description"] == "Existing schedule"
        assert schedule["ref"] == "scheduled/new-feature"
        assert schedule["cron"] == "0 */4 * * *"
        assert schedule["cron_timezone"] == "Stockholm"
        assert schedule["active"] is False

    def test__replace_existing_schedules(self, gitlab, group_and_project, schedules):

        replace_schedules = f"""
        projects_and_groups:
          {group_and_project}:
            schedules:
              "Existing schedule to replace":
                ref: scheduled/new-feature
                cron: "0 */3 * * *"
                cron_timezone: "London"
                active: true
        """

        run_gitlabform(replace_schedules, group_and_project)

        schedules = self.__find_pipeline_schedules_by_description(
            gitlab, group_and_project, "Existing schedule to replace"
        )
        assert schedules is not None
        assert len(schedules) == 1
        assert schedules[0]["description"] == "Existing schedule to replace"
        assert schedules[0]["ref"] == "scheduled/new-feature"
        assert schedules[0]["cron"] == "0 */3 * * *"
        assert schedules[0]["cron_timezone"] == "London"
        assert schedules[0]["active"] is True

    def test__delete_schedule(self, gitlab, group_and_project, schedules):

        delete_schedule = f"""
        projects_and_groups:
          {group_and_project}:
            schedules:
              "Redundant schedule":
                delete: True
        """

        run_gitlabform(delete_schedule, group_and_project)

        schedule = self.__find_pipeline_schedule_by_description_and_get_first(
            gitlab, group_and_project, "Redundant schedule"
        )
        assert schedule is None

    @staticmethod
    def __find_pipeline_schedule_by_description_and_get_first(
        gitlab, group_and_project, description
    ):
        for schedule in gitlab.get_all_pipeline_schedules(group_and_project):
            if schedule["description"] == description:
                return gitlab.get_pipeline_schedule(group_and_project, schedule["id"])
        return None

    @staticmethod
    def __find_pipeline_schedules_by_description(
        gitlab, group_and_project, description
    ):
        return list(
            map(
                lambda schedule: gitlab.get_pipeline_schedule(
                    group_and_project, schedule["id"]
                ),
                filter(
                    lambda schedule: schedule["description"] == description,
                    gitlab.get_all_pipeline_schedules(group_and_project),
                ),
            )
        )



from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.util.branch_protector import BranchProtector


class BranchesProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab, strict: bool):
        super().__init__("branches", gitlab)
        self.__branch_protector = BranchProtector(gitlab, strict)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        for branch in sorted(configuration["branches"]):
            self.__branch_protector.apply_branch_protection_configuration(
                project_and_group, configuration, branch
            )



import os
import pytest
import re

from gitlabform.gitlab import AccessLevel
from gitlabform.gitlab.core import NotFoundException
from tests.acceptance import run_gitlabform, DEFAULT_README


@pytest.fixture(scope="function")
def no_access_branch(request, gitlab, group_and_project):
    gitlab.create_branch(group_and_project, "no_access_branch", "main")
    gitlab.protect_branch(
        group_and_project,
        "no_access_branch",
        {
            "push_access_level": AccessLevel.NO_ACCESS.value,
            "merge_access_level": AccessLevel.NO_ACCESS.value,
            "unprotect_access_level": AccessLevel.MAINTAINER.value,
        },
    )

    def fin():
        gitlab.unprotect_branch(group_and_project, "no_access_branch")
        gitlab.delete_branch(group_and_project, "no_access_branch")

    request.addfinalizer(fin)


@pytest.fixture(scope="class")
def file(request):
    f = open("file.txt", "a")
    f.write("Hanzi (Simplified): 丏丅丙两\nHanzi (Traditional): 丕不丈丁")
    f.close()

    def fin():
        os.remove("file.txt")

    request.addfinalizer(fin)


@pytest.fixture(scope="class")
def file2(request):
    f = open("file2.txt", "a")
    f.write("LICENSE\n\n“This is a license file”\n")
    f.close()

    def fin():
        os.remove("file2.txt")

    request.addfinalizer(fin)


class TestFiles:
    def test__set_file_specific_branch(self, gitlab, group_and_project, branch):

        set_file_specific_branch = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                push_access_level: maintainer
                merge_access_level: developer
                unprotect_access_level: maintainer
            files:
              "README.md":
                overwrite: true
                branches:
                  - {branch}
                content: "Content for {branch} only"
        """

        run_gitlabform(set_file_specific_branch, group_and_project)

        commit = gitlab.get_last_commit(group_and_project, branch)
        assert all(
            [text in commit["message"] for text in ["Automated", "made by gitlabform"]]
        )

        file_content = gitlab.get_file(group_and_project, branch, "README.md")
        assert file_content == f"Content for {branch} only"

        other_branch_file_content = gitlab.get_file(
            group_and_project, "main", "README.md"
        )
        assert other_branch_file_content == DEFAULT_README

        # check if main stays protected after the file update
        the_branch = gitlab.get_branch(group_and_project, branch)
        assert the_branch["protected"] is True

    def test__set_file_strongly_protected_branch(
        self, gitlab, group_and_project, no_access_branch
    ):

        set_file_specific_branch = f"""
            projects_and_groups:
              {group_and_project}:
                branches:
                  no_access_branch:
                    protected: true
                    push_access_level: no access
                    merge_access_level: no access
                    unprotect_access_level: maintainer
                files:
                  "README.md":
                    overwrite: true
                    branches:
                      - no_access_branch
                    content: "Content for no_access_branch only"
            """

        run_gitlabform(set_file_specific_branch, group_and_project)

        commit = gitlab.get_last_commit(group_and_project, "no_access_branch")
        assert commit["message"] == "Automated change made by gitlabform"

        file_content = gitlab.get_file(
            group_and_project, "no_access_branch", "README.md"
        )
        assert file_content == "Content for no_access_branch only"

        other_branch_file_content = gitlab.get_file(
            group_and_project, "main", "README.md"
        )
        assert other_branch_file_content == DEFAULT_README

        # check if no_access_branch stays protected after the file update
        branch = gitlab.get_branch(group_and_project, "no_access_branch")
        assert branch["protected"] is True

    def test__delete_file_specific_branch(self, gitlab, group_and_project, branch):

        set_file_specific_branch = f"""
            projects_and_groups:
              {group_and_project}:
                branches:
                  {branch}:
                    protected: true
                    push_access_level: maintainer
                    merge_access_level: developer
                    unprotect_access_level: maintainer
                files:
                  "README.md":
                    branches:
                      - {branch}
                    delete: true
            """

        run_gitlabform(set_file_specific_branch, group_and_project)

        commit = gitlab.get_last_commit(group_and_project, branch)
        assert commit["message"] == "Automated delete made by gitlabform"

        with pytest.raises(NotFoundException):
            gitlab.get_file(group_and_project, branch, "README.md")

        # check if main stays protected after the file delete
        the_branch = gitlab.get_branch(group_and_project, branch)
        assert the_branch["protected"] is True

    def test__custom_commit_message(self, gitlab, group_and_project, branch):

        set_file_specific_branch = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: false
            files:
              "README.md":
                overwrite: true
                branches:
                  - {branch}
                skip_ci: true
                content: "Hello world!"
                commit_message: "Preconfigured commit message"
        """

        run_gitlabform(set_file_specific_branch, group_and_project)

        commit = gitlab.get_last_commit(group_and_project, branch)
        assert commit["message"] == "Preconfigured commit message [skip ci]"

    def test__set_file_protected_branches(self, gitlab, group_and_project, branch):

        test_config = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                push_access_level: {AccessLevel.MAINTAINER.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
            files:
              anyfile1:
                overwrite: true
                branches:
                  - {branch}
                skip_ci: true
                content: foobar
        """

        run_gitlabform(test_config, group_and_project)

        file_content = gitlab.get_file(group_and_project, branch, "anyfile1")
        assert file_content == "foobar"

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)
        assert push_access_levels == [AccessLevel.MAINTAINER.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

    def test__set_file_protected_branches_not_all_levels(
        self, gitlab, group_and_project, branch
    ):

        test_config = f"""
            projects_and_groups:
              {group_and_project}:
                branches:
                  {branch}:
                    protected: true
                    # use a level that will require unprotecting the branch
                    # to change the file value
                    push_access_level: {AccessLevel.MAINTAINER.value}
                    merge_access_level: {AccessLevel.MAINTAINER.value}
                files:
                  anyfile2:
                    overwrite: true
                    branches:
                      - {branch}
                    content: barfoo
            """

        run_gitlabform(test_config, group_and_project)

        file_content = gitlab.get_file(group_and_project, branch, "anyfile2")
        assert file_content == "barfoo"

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, branch)
        assert push_access_levels == [AccessLevel.MAINTAINER.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        # the default value
        # according to https://docs.gitlab.com/ee/api/protected_branches.html#protect-repository-branches
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

    def test__set_file_with_chinese_characters(self, gitlab, group_and_project, branch):
        set_file_chinese_characters = f"""
        projects_and_groups:
          {group_and_project}:
            files:
              "README.md":
                overwrite: true
                branches:
                  - {branch}
                content: |
                    Hanzi (Traditional): 丕不丈丁
                    Hanzi (Simplified): 丏丅丙两
        """

        run_gitlabform(set_file_chinese_characters, group_and_project)

        file_content = gitlab.get_file(group_and_project, branch, "README.md")
        assert file_content == "Hanzi (Traditional): 丕不丈丁\nHanzi (Simplified): 丏丅丙两\n"

    def test__set_external_file_with_chinese_characters(
        self, gitlab, group_and_project, branch, file
    ):
        set_file_chinese_characters = f"""
        projects_and_groups:
          {group_and_project}:
            files:
              "README.md":
                overwrite: true
                branches:
                  - {branch}
                file: file.txt
        """

        run_gitlabform(set_file_chinese_characters, group_and_project)

        file_content = gitlab.get_file(group_and_project, branch, "README.md")
        assert file_content == "Hanzi (Simplified): 丏丅丙两\nHanzi (Traditional): 丕不丈丁"

    def test__set_external_file_with_utf8_characters(
        self, gitlab, group_and_project, branch, file2
    ):
        set_file_chinese_characters = f"""
        projects_and_groups:
          {group_and_project}:
            files:
              "README.md":
                overwrite: true
                branches:
                  - {branch}
                file: file2.txt
        """

        run_gitlabform(set_file_chinese_characters, group_and_project)

        file_content = gitlab.get_file(group_and_project, branch, "README.md")
        assert file_content == "LICENSE\n\n“This is a license file”\n"



import functools
import os
from logging import debug
from urllib import parse

import pkg_resources
import requests

# noinspection PyPackageRequirements
import urllib3
from cli_ui import debug as verbose
from requests.adapters import HTTPAdapter

# noinspection PyPackageRequirements
from urllib3.util.retry import Retry

from gitlabform.configuration import Configuration
from gitlabform.util import to_str


class GitLabCore:
    def __init__(self, config_path=None, config_string=None):

        self.configuration = Configuration(config_path, config_string)

        self.url = self.configuration.get("gitlab|url", os.getenv("GITLAB_URL"))
        self.token = self.configuration.get("gitlab|token", os.getenv("GITLAB_TOKEN"))
        self.ssl_verify = self.configuration.get("gitlab|ssl_verify", True)
        self.timeout = self.configuration.get("gitlab|timeout", 10)

        self.session = requests.Session()

        retries = Retry(
            total=3, backoff_factor=0.25, status_forcelist=[500, 502, 503, 504]
        )

        self.session.mount("http://", HTTPAdapter(max_retries=retries))
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

        self.session.verify = self.ssl_verify
        if not self.ssl_verify:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        self.gitlabform_version = pkg_resources.get_distribution("gitlabform").version
        self.requests_version = pkg_resources.get_distribution("requests").version
        self.session.headers.update(
            {
                "private-token": self.token,
                "authorization": f"Bearer {self.token}",
                "user-agent": f"GitLabForm/{self.gitlabform_version} (python-requests/{self.requests_version})",
            }
        )

        try:
            version = self._make_requests_to_api("version")
            verbose(
                f"Connected to GitLab version: {version['version']} ({version['revision']})"
            )
            self.version = version["version"]
        except Exception as e:
            raise TestRequestFailedException(e)

    def get_configuration(self):
        return self.configuration

    def get_project(self, project_and_group_or_id):
        return self._make_requests_to_api("projects/%s", project_and_group_or_id)

    @functools.lru_cache()
    def _get_user_id(self, username: str) -> int:
        users = self._make_requests_to_api("users?username=%s", username, "GET")

        # this API endpoint is for lookup, not search, so 'username' has to be full and exact username
        # also it's not possible to get more than 1 user as a result

        if len(users) == 0:
            raise NotFoundException(
                "No users found when searching for username '%s'" % username
            )

        return int(users[0]["id"])

    @functools.lru_cache()
    def _get_group_id(self, path) -> int:
        group = self._make_requests_to_api("groups/%s", path, "GET")
        return int(group["id"])

    @functools.lru_cache()
    def _get_protected_branch_id(self, project_and_group_name, branch) -> int:
        branch = self._make_requests_to_api(
            "projects/%s/protected_branches/%s", (project_and_group_name, branch)
        )
        return int(branch["id"])

    @functools.lru_cache()
    def _get_project_id(self, project_and_group):
        # This is a NEW workaround for https://github.com/gitlabhq/gitlabhq/issues/8290
        result = self.get_project(project_and_group)
        return str(result["id"])

    def has_no_license(self):
        license = self._make_requests_to_api("license")
        return not license or license["expired"]

    def _make_requests_to_api(
        self,
        path_as_format_string,
        args=None,
        method="GET",
        data=None,
        expected_codes=200,
        json=None,
    ):
        """
        Makes an HTTP request or requests to the GitLab API endpoint. More than one request is made automatically
        if the endpoint is paginated. (See underlying method for authentication, retries, timeout etc.)

        :param path_as_format_string: path with parts to be replaced by values from `args` replaced by '%s'
                                      (aka the old-style Python string formatting, see:
                                       https://docs.python.org/2/library/stdtypes.html#string-formatting )
        :param args: single element or a tuple of values to put under '%s's in `path_as_format_string`
        :param method: uppercase string of a HTTP method name, like 'GET' or 'PUT'
        :param data: dict with data to be 'PUT'ted or 'POST'ed
        :param expected_codes: a single HTTP code (like: 200) or a list of accepted HTTP codes
                               - if the call to the API will return other code an exception will be thrown
        :param json: alternatively to `dict` you can set this to a string that can be parsed as JSON that will
                     be used as data to be 'PUT'ted or 'POST'ed
        :return: data returned by the endpoint, as a JSON object. If the API is paginated, it returns JSONs with
                 arrays of objects and then this method returns JSON with a single array that contains all of those
                 objects.
        """
        if method != "GET":
            response = self._make_request_to_api(
                path_as_format_string, args, method, data, expected_codes, json
            )
            return response.json()
        else:
            if "?" in path_as_format_string:
                path_as_format_string += "&per_page=100"
            else:
                path_as_format_string += "?per_page=100"

            first_response = self._make_request_to_api(
                path_as_format_string, args, method, data, expected_codes, json
            )
            results = first_response.json()

            # In newer versions of GitLab the 'x-total-pages' may not be available
            # anymore, see https://gitlab.com/gitlab-org/gitlab/-/merge_requests/43159
            # so let's use the 'x-next-page' header instead

            response = first_response
            while True:
                if (
                    "x-next-page" in response.headers
                    and response.headers["x-next-page"]
                ):
                    next_page = response.headers["x-next-page"]
                    response = self._make_request_to_api(
                        path_as_format_string + "&page=" + str(next_page),
                        args,
                        method,
                        data,
                        expected_codes,
                        json,
                    )
                    results += response.json()
                else:
                    break

        return results

    def _make_request_to_api(
        self, path_as_format_string, args, method, dict_data, expected_codes, json_data
    ):

        """
        Makes a single request to the GitLab API. Takes care of the authentication, basic error processing,
        retries, timeout etc.

        :param for the params description please see `_make_requests_to_api()`
        :return: data returned by the endpoint, as a JSON object.
        """

        expected_codes = self._listify(expected_codes)

        if dict_data and json_data:
            raise Exception(
                "You need to pass the data either as dict (dict_data) or JSON (json_data), not both!"
            )

        url = f"{self.url}/api/v4/{self._format_with_url_encoding(path_as_format_string, args)}"
        if dict_data:
            response = self.session.request(
                method, url, data=dict_data, timeout=self.timeout
            )
            debug(f"===> data = {to_str(dict_data)}")
        elif json_data:
            response = self.session.request(
                method, url, json=json_data, timeout=self.timeout
            )
            debug(f"===> json = {to_str(json_data)}")
        else:
            response = self.session.request(method, url, timeout=self.timeout)

        if response.status_code in expected_codes:
            # if we accept error responses then they will likely not contain a JSON body
            # so fake it to fix further calls to response.json()
            if response.status_code == 204 or (400 <= response.status_code <= 499):
                response.json = lambda: {}
        else:
            if response.status_code == 404:
                raise NotFoundException(
                    f"Resource with url='{url}' not found (HTTP 404)!"
                )
            else:
                if dict_data:
                    data_output = f"data='{to_str(dict_data)}' "
                elif json_data:
                    data_output = f"json='{to_str(json_data)}' "
                else:
                    data_output = ""

                raise UnexpectedResponseException(
                    f"Request url='{url}', method={method}, {data_output}failed -"
                    f" expected code(s) {str(expected_codes)},"
                    f" got code {response.status_code} & body: '{response.text}'",
                    response.status_code,
                    response.text,
                )
        if response.json():
            debug(f"<--- json = {to_str(response.json())}")
        else:
            debug(f"<--- json = (empty))")
        return response

    @staticmethod
    def _format_with_url_encoding(format_string, single_arg_or_args_tuple):

        # we want to URL-encode all the args, but not the path itself which looks like "/foo/%s/bar"
        # because '/'s here are NOT to be URL-encoded

        if not single_arg_or_args_tuple:
            # there are no params, so the format_string is the URL
            return format_string
        else:
            if type(single_arg_or_args_tuple) == tuple:
                # URL-encode each arg in the tuple and return it as tuple too
                url_encoded_args = ()
                for arg in single_arg_or_args_tuple:
                    url_encoded_args += (parse.quote_plus(str(arg)),)
            else:
                # URL-encode single arg
                url_encoded_args = parse.quote_plus(str(single_arg_or_args_tuple))

            return format_string % url_encoded_args

    @staticmethod
    def _listify(expected_codes):
        if isinstance(expected_codes, int):
            return [expected_codes]
        else:
            return expected_codes


class TestRequestFailedException(Exception):
    def __init__(self, underlying: Exception):
        self.underlying = underlying


class NotFoundException(Exception):
    pass


class TimeoutWaitingForDeletion(Exception):
    pass


class InvalidParametersException(Exception):
    pass


class UnexpectedResponseException(Exception):
    def __init__(self, message: str, response_status_code: int, response_text: str):
        self.message: str = message
        self.response_status_code: int = response_status_code
        self.response_text: str = response_text

    def __str__(self):
        return self.message



from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.gitlab.core import NotFoundException, UnexpectedResponseException
from gitlabform.processors.abstract_processor import AbstractProcessor


class ResourceGroupsProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("resource_groups", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        for config_resource_group_name in configuration["resource_groups"]:
            config_process_mode = configuration["resource_groups"][
                config_resource_group_name
            ]["process_mode"]

            try:
                gitlab_resource_group = self.gitlab.get_specific_resource_group(
                    project_and_group, config_resource_group_name
                )
            except NotFoundException:
                raise Exception(
                    f"Project is not configured to use resource group: {config_resource_group_name}.\n"
                    f"Add the resource group in your project's .gitlab-ci.yml file.\n"
                    f"For more information, visit https://docs.gitlab.com/ee/ci/resource_groups/#add-a-resource-group.",
                )

            # compare the resource group process mode between the config entity and gitlab entity
            if config_process_mode != gitlab_resource_group["process_mode"]:
                try:
                    self.gitlab.update_resource_group(
                        project_and_group,
                        config_resource_group_name,
                        {"process_mode": config_process_mode},
                    )
                except UnexpectedResponseException:
                    raise Exception(
                        f"process_mode does not have a valid value: {config_process_mode}"
                    )
                verbose(f"Setting resource group process mode to {config_process_mode}")



import pytest

from gitlabform.gitlab.core import UnexpectedResponseException, NotFoundException

from tests.acceptance import run_gitlabform


class TestVariables:
    def test__builds_disabled(self, gitlab, group_and_project):
        config_builds_not_enabled = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: disabled
            variables:
              foo:
                key: FOO
                value: 123
        """

        run_gitlabform(config_builds_not_enabled, group_and_project)

        with pytest.raises(UnexpectedResponseException):
            # variables will NOT be available without builds_access_level in ['private', 'enabled']
            gitlab.get_variables(group_and_project)

    def test__single_variable(self, gitlab, group_and_project):
        config_single_variable = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: enabled
            variables:
              foo:
                key: FOO
                value: 123
        """

        run_gitlabform(config_single_variable, group_and_project)

        variable = gitlab.get_variable(group_and_project, "FOO")
        assert variable == "123"

    def test__delete_variable(self, gitlab, group_and_project):
        config_single_variable = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: enabled
            variables:
              foo:
                key: FOO
                value: 123
        """

        run_gitlabform(config_single_variable, group_and_project)

        variable = gitlab.get_variable(group_and_project, "FOO")
        assert variable == "123"

        config_delete_variable = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: enabled
            variables:
              foo:
                key: FOO
                value: 123
                delete: true
        """

        run_gitlabform(config_delete_variable, group_and_project)

        with pytest.raises(NotFoundException):
            gitlab.get_variable(group_and_project, "FOO")

    def test__reset_single_variable(self, gitlab, group_and_project):
        config_single_variable = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: enabled
            variables:
              foo:
                key: FOO
                value: 123
        """

        run_gitlabform(config_single_variable, group_and_project)

        variable = gitlab.get_variable(group_and_project, "FOO")
        assert variable == "123"

        config_single_variable2 = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: enabled
            variables:
              foo:
                key: FOO
                value: 123456
        """

        run_gitlabform(config_single_variable2, group_and_project)

        variable = gitlab.get_variable(group_and_project, "FOO")
        assert variable == "123456"

    def test__more_variables(self, gitlab, group_and_project):
        config_more_variables = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: enabled
            variables:
              foo:
                key: FOO
                value: 123456
              bar:
                key: BAR
                value: bleble
        """

        run_gitlabform(config_more_variables, group_and_project)

        variable = gitlab.get_variable(group_and_project, "FOO")
        assert variable == "123456"
        variable = gitlab.get_variable(group_and_project, "BAR")
        assert variable == "bleble"

    def test__variable_with_env_scope(self, gitlab, group_and_project):
        config_more_variables = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: enabled
            variables:
              foo_ee:
                key: FOO1
                value: alfa
                environment_scope: test/ee
                filter[environment_scope]: test/ee
        """

        run_gitlabform(config_more_variables, group_and_project)

        variable = gitlab.get_variable(group_and_project, "FOO1", "test/ee")
        assert variable == "alfa"

    def test__variables_with_env_scope(self, gitlab, group_and_project):
        config_more_variables = f"""
        projects_and_groups:
          {group_and_project}:
            project_settings:
              builds_access_level: enabled
            variables:
              foo_ee:
                key: FOO2
                value: alfa
                environment_scope: test/ee
                filter[environment_scope]: test/ee
              foo_lv:
                key: FOO2
                value: beta
                environment_scope: test/lv
                filter[environment_scope]: test/lv
        """

        run_gitlabform(config_more_variables, group_and_project)

        variables = gitlab.get_variables(group_and_project)
        variables_keys = [variable["key"] for variable in variables]
        assert "FOO2" in variables_keys

        variable = gitlab.get_variable(group_and_project, "FOO2", "test/ee")
        assert variable == "alfa"
        variable = gitlab.get_variable(group_and_project, "FOO2", "test/lv")
        assert variable == "beta"



from tests.acceptance import run_gitlabform


class TestFilesAll:

    # this test should be in a separate class than other test files as it changes too
    # much for a reasonable setup and cleanup using fixtures
    def test__set_file_all_branches(
        self, gitlab, group_and_project, branch, other_branch
    ):
        set_file_all_branches = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              {branch}:
                protected: true
                push_access_level: maintainer
                merge_access_level: developer
                unprotect_access_level: maintainer
              {other_branch}:
                protected: false
            files:
              "README.md":
                overwrite: true
                branches: all
                content: "Content for all branches"
        """
        run_gitlabform(set_file_all_branches, group_and_project)

        for some_branch in [
            "main",
            branch,
            other_branch,
        ]:
            file_content = gitlab.get_file(group_and_project, some_branch, "README.md")
            assert file_content == "Content for all branches"

        # check that this branch remains unprotected
        for some_branch in [
            other_branch,
        ]:
            some_branch = gitlab.get_branch(group_and_project, some_branch)
            assert some_branch["protected"] is False



from gitlabform.gitlab import GitLab
from gitlabform.processors.defining_keys import Key, And
from gitlabform.processors.multiple_entities_processor import MultipleEntitiesProcessor


class ProtectedEnvironmentsProcessor(MultipleEntitiesProcessor):
    """https://docs.gitlab.com/ee/api/protected_environments.html#protect-repository-environments"""

    def __init__(self, gitlab: GitLab):
        super().__init__(
            "protected_environments",
            gitlab,
            list_method_name=gitlab.list_protected_environments,
            add_method_name=gitlab.protect_a_repository_environment,
            delete_method_name=gitlab.unprotect_environment,
            defining=Key("name"),
            required_to_create_or_update=And(Key("name"), Key("deploy_access_levels")),
        )

        self.custom_diff_analyzers[
            "deploy_access_levels"
        ] = self.recursive_diff_analyzer



from abc import ABC

import functools
from logging import debug

from gitlabform.configuration import ConfigurationGroups
from gitlabform.util import to_str


class ConfigurationProjects(ConfigurationGroups, ABC):
    """
    Gets the projects, skipped projects and their effective configuration.

    ConfigurationGroups is an ancestor of this class because the effective project configuration
    depends on the groups (and common) configuration.
    """

    def get_projects(self) -> list:
        """
        :return: sorted list of projects names, that are EXPLICITLY defined in the config
        """
        projects = []
        projects_and_groups = self.get("projects_and_groups")
        for element in projects_and_groups.keys():
            if element != "*" and not element.endswith("/*"):
                projects.append(element)
        return sorted(projects)

    def is_project_skipped(self, project) -> bool:
        """
        :return: if project is defined in the key with projects to skip,
                 ignoring the case
        """
        return self._is_skipped_case_insensitively(
            self.get("skip_projects", []), project
        )

    @functools.lru_cache()
    def get_effective_config_for_project(self, group_and_project) -> dict:
        """
        :param group_and_project: "project_group/project_name"
        :return: merged configuration for this project, from common, group/subgroup and project level.
                 If project belongs to a subgroup, like "x/y/z", then it gets config from both group "x" as well
                 as subgroup "y".
                 Merging is additive.
        """

        group, _ = group_and_project.rsplit("/", 1)

        effective_config_for_group = self.get_effective_config_for_group(group)

        project_config = self._get_project_config(group_and_project)
        debug("Project config: %s", to_str(project_config))

        if not effective_config_for_group and project_config:
            self._validate_break_inheritance_flag(project_config, group_and_project)

        effective_config_for_project = self._merge_configs(
            effective_config_for_group,
            project_config,
        )
        debug(
            "*Effective* config common+group/subgroup+project: %s",
            to_str(effective_config_for_project),
        )

        return effective_config_for_project

    def _get_project_config(self, group_and_project) -> dict:
        """
        :param group_and_project: 'group/project'
        :return: configuration for this project or empty dict if not defined,
                 ignoring the case
        """
        return self._get_case_insensitively(
            self.get("projects_and_groups"), group_and_project
        )



from logging import debug
from cli_ui import debug as verbose
from cli_ui import fatal

from gitlabform.constants import EXIT_INVALID_INPUT
from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor


class MembersProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("members", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):

        enforce_members = configuration.get("members|enforce", False)

        groups = configuration.get("members|groups", {})

        users = configuration.get("members|users", {})

        if not groups and not users and not enforce_members:
            fatal(
                "Project members configuration section has to contain"
                " either 'users' or 'groups' non-empty keys"
                " (unless you want to enforce no direct members).",
                exit_code=EXIT_INVALID_INPUT,
            )

        self._process_groups(project_and_group, groups, enforce_members)
        self._process_users(project_and_group, users, enforce_members)

    def _process_groups(
        self, project_and_group: str, groups: dict, enforce_members: bool
    ):
        if groups:

            verbose("Processing groups as members...")

            current_groups = self.gitlab.get_groups_from_project(project_and_group)
            for group in groups:
                expires_at = (
                    groups[group]["expires_at"].strftime("%Y-%m-%d")
                    if "expires_at" in groups[group]
                    else None
                )
                access_level = (
                    groups[group]["group_access"]
                    if "group_access" in groups[group]
                    else None
                )

                # we only add the group if it doesn't have the correct settings
                if (
                    group in current_groups
                    and expires_at == current_groups[group]["expires_at"]
                    and access_level == current_groups[group]["group_access_level"]
                ):
                    debug("Ignoring group '%s' as it is already a member", group)
                    debug(
                        "Current settings for '%s' are: %s"
                        % (group, current_groups[group])
                    )
                else:
                    debug("Setting group '%s' as a member", group)
                    access = access_level
                    expiry = expires_at

                    # we will remove group access first and then re-add them,
                    # to ensure that the groups have the expected access level
                    self.gitlab.unshare_with_group(project_and_group, group)
                    self.gitlab.share_with_group(
                        project_and_group, group, access, expiry
                    )

        if enforce_members:

            current_groups = self.gitlab.get_groups_from_project(project_and_group)

            groups_in_config = groups.keys()
            groups_in_gitlab = current_groups.keys()
            groups_not_in_config = set(groups_in_gitlab) - set(groups_in_config)

            for group_not_in_config in groups_not_in_config:
                debug(
                    f"Removing group '{group_not_in_config}' that is not configured to be a member."
                )
                self.gitlab.unshare_with_group(project_and_group, group_not_in_config)
        else:
            debug("Not enforcing group members.")

    def _process_users(
        self, project_and_group: str, users: dict, enforce_members: bool
    ):
        if users:

            verbose("Processing users as members...")

            current_members = self.gitlab.get_members_from_project(project_and_group)
            for user in users:
                expires_at = (
                    users[user]["expires_at"].strftime("%Y-%m-%d")
                    if "expires_at" in users[user]
                    else None
                )
                access_level = (
                    users[user]["access_level"]
                    if "access_level" in users[user]
                    else None
                )
                # we only add the user if it doesn't have the correct settings
                if (
                    user in current_members
                    and expires_at == current_members[user]["expires_at"]
                    and access_level == current_members[user]["access_level"]
                ):
                    debug("Ignoring user '%s' as it is already a member", user)
                    debug(
                        "Current settings for '%s' are: %s"
                        % (user, current_members[user])
                    )
                else:
                    debug("Setting user '%s' as a member", user)
                    access = access_level
                    expiry = expires_at
                    self.gitlab.remove_member_from_project(project_and_group, user)
                    self.gitlab.add_member_to_project(
                        project_and_group, user, access, expiry
                    )

        if enforce_members:

            current_members = self.gitlab.get_members_from_project(project_and_group)

            users_in_config = users.keys()
            users_in_gitlab = current_members.keys()
            users_not_in_config = set(users_in_gitlab) - set(users_in_config)

            for user_not_in_config in users_not_in_config:
                debug(
                    f"Removing user '{user_not_in_config}' that is not configured to be a member."
                )
                self.gitlab.remove_member_from_project(
                    project_and_group, user_not_in_config
                )
        else:
            debug("Not enforcing user members.")



from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor


class IntegrationsProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("integrations", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        for integration in sorted(configuration["integrations"]):
            if configuration.get("integrations|" + integration + "|delete"):
                verbose(f"Deleting integration: {integration}")
                self.gitlab.delete_integration(project_and_group, integration)
            else:
                verbose(f"Setting integration: {integration}")
                self.gitlab.set_integration(
                    project_and_group,
                    integration,
                    configuration["integrations"][integration],
                )



from cli_ui import debug as verbose
from cli_ui import warning

import copy
import textwrap
import ez_yaml

from gitlabform.gitlab import GitLab
from gitlabform.processors.util.difference_logger import hide
from gitlabform.processors.defining_keys import Key, And, OptionalKey
from gitlabform.processors.multiple_entities_processor import MultipleEntitiesProcessor


class VariablesProcessor(MultipleEntitiesProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__(
            "variables",
            gitlab,
            list_method_name="get_variables",
            add_method_name="post_variable",
            delete_method_name="delete_variable",
            defining=And(Key("key"), OptionalKey("environment_scope")),
            required_to_create_or_update=And(Key("key"), Key("value")),
            edit_method_name="put_variable",
        )

    def _can_proceed(self, project_or_group: str, configuration: dict):
        if (
            self.gitlab.get_project_settings(project_or_group).get(
                "builds_access_level"
            )
            == "disabled"
        ):
            warning("Builds disabled in this project so I can't set variables here.")
            return False
        else:
            return True

    def _print_diff(self, project_and_group: str, configuration):

        try:
            current_variables = self.gitlab.get_variables(project_and_group)

            for variable in current_variables:
                variable["value"] = hide(variable["value"])

            verbose(f"Variables for {project_and_group} in GitLab:")

            verbose(
                textwrap.indent(
                    ez_yaml.to_string(current_variables),
                    "  ",
                )
            )
        except:
            verbose(f"Variables for {project_and_group} in GitLab cannot be checked.")

        verbose(f"Variables in {project_and_group} in configuration:")

        configured_variables = copy.deepcopy(configuration)
        for key in configured_variables.keys():
            configured_variables[key]["value"] = hide(
                configured_variables[key]["value"]
            )

        verbose(
            textwrap.indent(
                ez_yaml.to_string(configured_variables),
                "  ",
            )
        )



from abc import ABC

from cli_ui import debug as verbose

from typing import List

from gitlabform.configuration import Configuration
from gitlabform.gitlab import GitLab
from gitlabform.output import EffectiveConfigurationFile
from gitlabform.processors.abstract_processor import AbstractProcessor


class AbstractProcessors(ABC):
    def __init__(self, gitlab: GitLab, config: Configuration, strict: bool):
        self.processors: List[AbstractProcessor] = []

    def get_configuration_names(self):
        return [processor.configuration_name for processor in self.processors]

    def process_entity(
        self,
        entity_reference: str,
        configuration: dict,
        dry_run: bool,
        effective_configuration: EffectiveConfigurationFile,
        only_sections: List[str],
    ):
        for processor in self.processors:
            if only_sections == "all" or processor.configuration_name in only_sections:
                processor.process(
                    entity_reference, configuration, dry_run, effective_configuration
                )
            else:
                verbose(
                    f"Skipping section '{processor.configuration_name}' - not in --only-sections list."
                )



from functools import wraps


class SafeDict(dict):
    """
    A dict that a "get" method that allows to use a path-like reference to its subdict values.

    For example with a dict like {"key": {"subkey": {"subsubkey": "value"}}}
    you can use a string 'key|subkey|subsubkey' to get the 'value'.

    The default value is returned if ANY of the subelements does not exist.

    Code based on https://stackoverflow.com/a/44859638/2693875
    """

    def get(self, path, default=None):
        keys = path.split("|")
        val = None

        for key in keys:
            if val:
                if isinstance(val, list):
                    val = [v.get(key, default) if v else None for v in val]
                else:
                    val = val.get(key, default)
            else:
                val = dict.get(self, key, default)

            if not val:
                break

        return val


def configuration_to_safe_dict(method):
    """
    This wrapper function calls the method with the configuration converted from a regular dict into a SafeDict
    """

    @wraps(method)
    def method_wrapper(self, project_and_group, configuration, *args):
        return method(self, project_and_group, SafeDict(configuration), *args)

    return method_wrapper
